"""Create dummy datasets.

The datasets will have the same directory and files structure as the original ones, but only contain a single sample
for each directory. The samples are just randomly generated noise.

The main purpose of this script is to be used with tests. But it can also be useful to visualize the structure of a dataset
without having to download it.
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
from pathlib import Path
from typing import Tuple, Union

import cv2 as cv
import numpy as np

from ptlflow.utils import flow_utils
from ptlflow.utils.utils import config_logging

config_logging()


def write_flying_chairs(
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (384, 512)
) -> None:
    """Generate a dummy version of the Flying Chairs dataset.

    The original dataset is available at:

    https://lmb.informatik.uni-freiburg.de/resources/datasets/FlyingChairs.en.htm

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (384, 512)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size+(3,))
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)

    root_dir = Path(root_dir) / 'FlyingChairs_release'
    data_dir = root_dir / 'data'
    data_dir.mkdir(parents=True, exist_ok=True)

    cv.imwrite(str(data_dir / '00001_img1.ppm'), img)
    cv.imwrite(str(data_dir / '00001_img2.ppm'), img)
    flow_utils.flow_write(data_dir / '00001_flow.flo', flow)

    logging.info('Created dataset on %s.', str(root_dir))


def write_flying_chairs2(
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (384, 512)
) -> None:
    """Generate a dummy version of the Flying Chairs 2 dataset.

    The original dataset is available at:

    https://lmb.informatik.uni-freiburg.de/resources/datasets/FlyingChairs.en.html

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (384, 512)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size+(3,), np.uint8)
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)
    mask = np.random.randint(0, 2, img_size, np.uint8) * 255

    root_dir = Path(root_dir) / 'FlyingChairs2'
    for split in ['train', 'val']:
        img_dir = root_dir / split
        img_dir.mkdir(parents=True, exist_ok=True)

        cv.imwrite(str(img_dir / '0000001-img_0.png'), img)
        cv.imwrite(str(img_dir / '0000001-img_1.png'), img)
        cv.imwrite(str(img_dir / '0000001-oids_0.png'), img)
        cv.imwrite(str(img_dir / '0000001-oids_1.png'), img)
        cv.imwrite(str(img_dir / '0000001-occ_01.png'), mask)
        cv.imwrite(str(img_dir / '0000001-occ_10.png'), mask)
        cv.imwrite(str(img_dir / '0000001-mb_01.png'), mask)
        cv.imwrite(str(img_dir / '0000001-mb_10.png'), mask)
        flow_utils.flow_write(img_dir / '0000001-flow_01.flo', flow)
        flow_utils.flow_write(img_dir / '0000001-flow_10.flo', flow)

    logging.info('Created dataset on %s.', str(root_dir))


def write_hd1k(
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (1080, 2560)
) -> None:
    """Generate a dummy version of the HD1K dataset.

    The original dataset is available at:

    http://hci-benchmark.org/

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (1080, 2560)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size)
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)

    root_dir = Path(root_dir) / 'HD1K'
    for dir in ['hd1k_challenge', 'hd1k_input']:
        img_dir = root_dir / dir / 'image_2'
        img_dir.mkdir(parents=True, exist_ok=True)
        cv.imwrite(str(img_dir / '000000_0010.png'), img)
        cv.imwrite(str(img_dir / '000000_0011.png'), img)

    flow_path = root_dir / 'hd1k_flow_gt' / 'flow_occ' / '000000_0010.png'
    flow_path.parent.mkdir(parents=True, exist_ok=True)
    flow_utils.flow_write(flow_path, flow)

    unc_path = root_dir / 'hd1k_flow_uncertainty' / 'flow_unc' / '000000_0010.png'
    unc_path.parent.mkdir(parents=True, exist_ok=True)
    cv.imwrite(str(unc_path), img)

    logging.info('Created dataset on %s.', str(root_dir))


def write_kitti(
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (375, 1242)
) -> None:
    """Generate a dummy version of the KITTI 2012 and 2015 datasets.

    The original datasets are available at:

    http://www.cvlibs.net/datasets/kitti/eval_stereo_flow.php?benchmark=flow

    http://www.cvlibs.net/datasets/kitti/eval_scene_flow.php?benchmark=flow

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (375, 1242)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size)
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)

    root_dir = Path(root_dir) / 'KITTI'
    for year in ['2012', '2015']:
        for split in ['training', 'testing']:
            if year == '2012':
                img_dir = 'colored_0'
            else:
                img_dir = 'image_2'

            img_dir_path = root_dir / year / split / img_dir
            img_dir_path.mkdir(parents=True, exist_ok=True)
            cv.imwrite(str(img_dir_path / '000000_10.png'), img)
            cv.imwrite(str(img_dir_path / '000000_11.png'), img)

            if split == 'training':
                flow_path = root_dir / year / split / 'flow_occ' / '000000_10.png'
                flow_path.parent.mkdir(parents=True, exist_ok=True)
                flow_utils.flow_write(flow_path, flow)

    logging.info('Created dataset on %s.', str(root_dir))


def write_sintel(
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (436, 1024)
) -> None:
    """Generate a dummy version of the MPI Sintel dataset.

    The original dataset is available at:

    http://sintel.is.tue.mpg.de/

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (436, 1024)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size)
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)
    mask = np.random.randint(0, 2, img_size, np.uint8) * 255

    root_dir = Path(root_dir) / 'MPI-Sintel'
    for split in ['training', 'test']:
        for pass_name in ['clean', 'final']:
            img_dir_path = root_dir / split / pass_name / 'sequence_1'
            img_dir_path.mkdir(parents=True, exist_ok=True)
            cv.imwrite(str(img_dir_path / 'frame_0001.png'), img)
            cv.imwrite(str(img_dir_path / 'frame_0002.png'), img)

        if split == 'training':
            flow_path = root_dir / split / 'flow' / 'sequence_1' / 'frame_0001.flo'
            flow_path.parent.mkdir(parents=True, exist_ok=True)
            flow_utils.flow_write(flow_path, flow)

            occ_path = root_dir / split / 'occlusions' / 'sequence_1' / 'frame_0001.png'
            occ_path.parent.mkdir(parents=True, exist_ok=True)
            cv.imwrite(str(occ_path), mask)

    logging.info('Created dataset on %s.', str(root_dir))


def write_things(  # noqa: C901
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (540, 960)
) -> None:
    """Generate a dummy version of the Flying Things 3D dataset.

    The original dataset is available at:

    https://lmb.informatik.uni-freiburg.de/resources/datasets/SceneFlowDatasets.en.html

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (540, 960)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size)
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)
    mask = np.random.randint(0, 2, img_size, np.uint8) * 255

    root_dir = Path(root_dir) / 'FlyingThings3D'
    for cat in ['frames_cleanpass', 'frames_finalpass', 'optical_flow', 'occlusions', 'motion_boundaries']:
        for split in ['TEST', 'TRAIN']:
            for letter in ['A', 'B', 'C']:
                for side_dir, side_name in [('left', 'L'), ('right', 'R')]:
                    if cat == 'optical_flow':
                        for direc_dir, direc_name in [('into_future', 'IntoFuture'), ('into_past', 'IntoPast')]:
                            flow_dir = root_dir / cat / split / letter / '0000' / direc_dir / side_dir
                            flow_dir.mkdir(parents=True, exist_ok=True)
                            for num in range(2):
                                flow_utils.flow_write(
                                    flow_dir / f'OpticalFlow{direc_name}_{num:04d}_{side_name}.pfm', flow)
                    elif cat.startswith('frames'):
                        img_dir_path = root_dir / cat / split / letter / '0000' / side_dir
                        img_dir_path.mkdir(parents=True, exist_ok=True)
                        cv.imwrite(str(img_dir_path / '0000.png'), img)
                        cv.imwrite(str(img_dir_path / '0001.png'), img)
                    else:
                        for direc_dir, direc_name in [('into_future', 'IntoFuture'), ('into_past', 'IntoPast')]:
                            mask_dir = root_dir / cat / split / letter / '0000' / direc_dir / side_dir
                            mask_dir.mkdir(parents=True, exist_ok=True)
                            for num in range(2):
                                cv.imwrite(str(mask_dir / f'{cat}{direc_name}_{num:04d}_{side_name}.png'), mask)

    logging.info('Created dataset on %s.', str(root_dir))


def write_things_subset(
    root_dir: Union[str, Path],
    img_size: Tuple[int, int] = (540, 960)
) -> None:
    """Generate a dummy version of the Flying Things 3D subset dataset.

    The original dataset is available at:

    https://lmb.informatik.uni-freiburg.de/resources/datasets/SceneFlowDatasets.en.html

    Parameters
    ----------
    root_dir : Union[str, Path]
        Path to the directory where the dummy dataset will be created.
    img_size : Tuple[int, int], default (540, 960)
        The size of the images inside of this dataset.
    """
    img = np.random.randint(0, 256, img_size)
    flow = np.random.rand(img_size[0], img_size[1], 2).astype(np.float32)
    mask = np.random.randint(0, 2, img_size, np.uint8) * 255

    root_dir = Path(root_dir) / 'FlyingThings3D_subset'
    for split in ['train', 'val']:
        for cat in ['image_clean', 'image_final', 'flow', 'flow_occlusions', 'motion_boundaries']:
            for side in ['left', 'right']:
                if cat == 'image_clean' or cat == 'image_final':
                    img_dir_path = root_dir / split / cat / side
                    img_dir_path.mkdir(parents=True, exist_ok=True)
                    cv.imwrite(str(img_dir_path / '0000000.png'), img)
                    cv.imwrite(str(img_dir_path / '0000001.png'), img)
                else:
                    for direc in ['into_future', 'into_past']:
                        if direc == 'into_future':
                            num = '0000000'
                        else:
                            num = '0000001'

                        if cat == 'flow':
                            flow_path = root_dir / split / cat / side / direc / f'{num}.flo'
                            flow_path.parent.mkdir(parents=True, exist_ok=True)
                            flow_utils.flow_write(flow_path, flow)
                        else:
                            img_path = root_dir / split / cat / side / direc / f'{num}.png'
                            img_path.parent.mkdir(parents=True, exist_ok=True)
                            cv.imwrite(str(img_path), mask)

    logging.info('Created dataset on %s.', str(root_dir))



from argparse import ArgumentParser
from pathlib import Path
import random

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import IterSpatialCorrelationSampler as SpatialCorrelationSampler
import torch
import torch.nn as nn

from .pwc_modules import conv, upsample2d_as, rescale_flow, initialize_msra
from .pwc_modules import WarpingLayer, FeatureExtractor, ContextNetwork,\
    FlowEstimatorDense, OccContextNetwork, OccEstimatorDense
from .irr_modules import OccUpsampleNetwork, RefineFlow, RefineOcc
from ..base_model.base_model import BaseModel
from .losses import MultiScaleEPE_PWC_Bi_Occ_upsample


class ScopeFlow(BaseModel):
    pretrained_checkpoints = {
        'chairs': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-chairs-ebfaa62d.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-things-70e22d63.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-kitti-a20c434d.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-sintel-21a91683.ckpt'
    }

    def __init__(self, args):
        super(ScopeFlow, self).__init__(
            args=args,
            loss_fn=MultiScaleEPE_PWC_Bi_Occ_upsample(args),
            output_stride=64)
        self.pwc_groups = ['extractor', '1x1',
                           'flow_estimator',
                           'context', 'refine_flow',
                           'attention'] + ['occ'] * 3

        # Init sizes
        self.args = args
        self.search_range = 4
        self.num_chs = [3, 16, 32, 64, 96, 128, 196]
        self.output_level = 4
        self.num_levels = 7
        self.ch_proj_size = 32
        self.min_layer = len(self.num_chs) - self.output_level - 1

        # Init generic layers
        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)
        self.feature_pyramid_extractor = FeatureExtractor(self.num_chs)
        self.warping_layer = WarpingLayer()
        self.correlation = SpatialCorrelationSampler(kernel_size=1, patch_size=2*self.args.search_range+1, padding=0)

        # Calc dimensions
        self.dim_corr = (self.search_range * 2 + 1) ** 2
        self.num_ch_in_flo = self.dim_corr + self.ch_proj_size + 2
        self.num_ch_in_occ = self.dim_corr + self.ch_proj_size + 1

        # Init PWC modules
        self.flow_estimators = FlowEstimatorDense(self.num_ch_in_flo)
        self.context_networks = ContextNetwork(self.num_ch_in_flo + 448 + 2)
        self.occ_estimators = OccEstimatorDense(self.num_ch_in_occ)
        self.occ_context_networks = OccContextNetwork(self.num_ch_in_occ + 448 + 1)
        self.occ_shuffle_upsample = OccUpsampleNetwork(11, 1)
        if hasattr(args, 'dropout') and self.args.dropout:
            print("Starting dropout!")
            self.dropout = nn.Dropout2d(p=self.args.dropout, inplace=False)

        self.conv_1x1 = nn.ModuleList([
            conv(self.num_chs[ch], self.ch_proj_size, kernel_size=1, stride=1, dilation=1)
            for ch in range(len(self.num_chs) - 1, self.min_layer, -1)])

        self.conv_1x1_1 = conv(self.num_chs[self.min_layer - 1], 3, kernel_size=1, stride=1, dilation=1)

        # Initialize attention
        if hasattr(args, 'self_attention') and self.args.self_attention:
            print("Initializing self attention!")
            from models.self_attention import AttentionConv, AttentionStem, SelfAttentionConv2d
            self.attention_conv = SelfAttentionConv2d(2, 2, kernel_size=5, padding=2)
        self.refine_flow = RefineFlow(2 + 1 + self.ch_proj_size, args.bilateral_mask)
        self.refine_occ = RefineOcc(1 + self.ch_proj_size + self.ch_proj_size)

        # Init weights
        initialize_msra(self.modules())
        self.param_groups = self._get_param_groups(self.pwc_groups)

        if hasattr(args, 'freeze_list') and args.freeze_list:
            self.freezed_params = args.freeze_list.split(',')
            self._freeze()
            self.random_freeze = False
        elif hasattr(args, 'random_freeze') and args.random_freeze:
            self.random_freeze = True
        self.mean_per_module = None

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=0.05)
        parser.add_argument('--search_range', type=int, default=4)
        parser.add_argument('--output_level', type=int, default=4)
        parser.add_argument('--num_levels', type=int, default=7)
        parser.add_argument('--num_chs', type=int, nargs='+', default=[3, 16, 32, 64, 96, 128, 196])
        parser.add_argument('--dropout', type=float, default=0.0)
        parser.add_argument('--self_attention', action='store_true')
        parser.add_argument('--bilateral_mask', action='store_true')
        parser.add_argument('--random_freeze', action='store_true')
        parser.add_argument('--seploss', action='store_true')
        parser.add_argument('--loss_perc', action='store_true')
        return parser

    def forward(self, inputs):

        # Extract input
        x1_raw = inputs['images'][:, 0]
        x2_raw = inputs['images'][:, 1]
        batch_size, _, height_im, width_im = x1_raw.size()

        # Get pyramid, on the bottom level are original images
        cur_x1 = self.feature_pyramid_extractor(x1_raw)
        cur_x2 = self.feature_pyramid_extractor(x2_raw)
        if hasattr(self.args, 'dropout') and self.args.dropout:
            # print("Dropout in ctx!")
            for xl in range(len(cur_x1)):
                # cur_x1[xl] = self.dropout(cur_x1[xl])
                cur_x2[xl] = self.dropout(cur_x2[xl])
        x1_pyramid = cur_x1 + [x1_raw]
        x2_pyramid = cur_x2 + [x2_raw]

        # Set output data structures
        output_dict = {}
        output_dict_eval = {}
        flows = []
        occs = []

        # Pre allocate output tensors
        flow_f, flow_b, occ_f, occ_b = self._allocate_out_tensors(x1_pyramid, batch_size)

        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):

            if l <= self.output_level:
                # Warp
                x1_warp, x2_warp, flow_f, flow_b, occ_f, occ_b = self._warp(l, x1, x2,
                                                                            flow_f,
                                                                            flow_b, occ_f,
                                                                            occ_b,
                                                                            height_im,
                                                                            width_im)

                # Correlate
                out_corr_relu_f = self._correlate(x1, x2_warp)
                out_corr_relu_b = self._correlate(x2, x1_warp)

                # Squash to channels projection size
                x1_1by1, x2_1by1 = self._squash(l, x1, x2)

                # Rescale Flow
                flow_f, flow_b = self._rescale_flow(flow_f, flow_b, height_im, width_im)

                # Estimate flow
                flow_cont_f, flow_cont_b = self._estimate_flow(out_corr_relu_f,
                                                               out_corr_relu_b,
                                                               x1_1by1, x2_1by1,
                                                               flow_f, flow_b)

                # Estimate occlusions
                occ_cont_f, occ_cont_b = self._estimate_occ(out_corr_relu_f,
                                                            out_corr_relu_b,
                                                            x1_1by1, x2_1by1,
                                                            occ_f, occ_b)

                # Prepare refinement inputs
                img1_resize, img2_resize, img1_warp, img2_warp, flow_cont_f, flow_cont_b = \
                    self._resize_and_warp_inputs(
                    x1_raw, x2_raw, flow_f, flow_b, flow_cont_f, flow_cont_b, height_im,
                    width_im)

                # Refine Flow
                flow_f, flow_cont_f, flow_b, flow_cont_b = self._refine_flow(
                    flow_cont_f, flow_cont_b, img1_resize, img2_resize, img1_warp,
                    img2_warp, x1_1by1, x2_1by1, height_im, width_im)

                # Refine Occlusions
                occ_f, occ_cont_f, occ_b, occ_cont_f = self._refine_occ(
                    x1_1by1, x2_1by1, flow_f, flow_b, occ_cont_f, occ_cont_b,
                    height_im, width_im)

                # Collect layer's outputs
                flows.append([flow_cont_f, flow_cont_b, flow_f, flow_b])
                occs.append([occ_cont_f, occ_cont_b, occ_f, occ_b])

            else:
                # Final flow upsampling
                flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                flow_b = upsample2d_as(flow_b, x2, mode="bilinear")

                # Final occ upsampling
                occ_f, occ_b = self.occ_upsampling(l, x1, x2, flow_f, flow_b,
                                                   occ_f, occ_b, height_im, width_im)

                # Apply self attention
                if hasattr(self.args, 'self_attention') and self.args.self_attention:
                    flow_f = self.attention_conv(flow_f)
                    flow_b = self.attention_conv(flow_b)

                # Aggregate outputs
                flows.append([flow_f, flow_b])
                occs.append([occ_f, occ_b])

        outputs = {}
        if self.training:
            outputs['flow_preds'] = flows
            outputs['occ_preds'] = occs
            outputs['flows'] = (upsample2d_as(flow_f, x1_raw, mode="bilinear") / self.args.div_flow)[:, None]
            outputs['occs'] = upsample2d_as(torch.sigmoid(occ_f), x1_raw, mode="bilinear")[:, None]
            outputs['flows_b'] = (upsample2d_as(flow_b, x1_raw, mode="bilinear") / self.args.div_flow)[:, None]
            outputs['occs_b'] = upsample2d_as(torch.sigmoid(occ_b), x1_raw, mode="bilinear")[:, None]
        else:
            outputs['flows'] = (upsample2d_as(flow_f, x1_raw, mode="bilinear") / self.args.div_flow)[:, None]
            outputs['occs'] = upsample2d_as(torch.sigmoid(occ_f), x1_raw, mode="bilinear")[:, None]
            outputs['flows_b'] = (upsample2d_as(flow_b, x1_raw, mode="bilinear") / self.args.div_flow)[:, None]
            outputs['occs_b'] = upsample2d_as(torch.sigmoid(occ_b), x1_raw, mode="bilinear")[:, None]
        return outputs

    def submodules_summary(self):
        print("Trainable submodules: {}".format({p[0] for p in self.named_parameters()
                                                 if p[1].requires_grad}))
        if not self.mean_per_module:
            print("Initializing mean per module")
            self.mean_per_module = {p[0]:p[1].data.mean()
                                    for p in self.named_parameters()}
        else:

            changes = {p[0]:p[1].data.mean() for p in self.named_parameters()
                       if p[1].data.mean() != self.mean_per_module[p[0]]}
            self.mean_per_module = {p[0]: p[1].data.mean() for p in
                                    self.named_parameters()}
            # print("Mean per submodule: {}".format(self.mean_per_module))
            print(f"Changes: {changes}")

    def freeze_random_weights(self):
        self.freezed_params = random.sample(self.pwc_groups,
                                            random.randrange(1, len(self.pwc_groups) / 2))
        self._freeze(verify=True)

    def _freeze(self, verify=False):
        print(f"Freezing groups {self.freezed_params}")

        if verify:
            for param in self.named_parameters():
                if not param[1].requires_grad:
                    param[1].requires_grad = True

        fkeys = set()
        for freezing_group in self.freezed_params:
            for key, param in self.param_groups[freezing_group].items():
                param.requires_grad = False
                if param.grad is not None:
                    param.grad.data.zero_()
                    # print("Freezing {}, grad sum is {}".format(key, param.grad.sum()))
                fkeys.add(key)

        assert len({p[0] for p in self.named_parameters() if not p[1].requires_grad}) == \
               len(fkeys)
        print(f"Freezed keys {fkeys}")
        print("Params with grad {}".format({p[0] for p in self.named_parameters()
                                            if p[1].requires_grad}))

    def _get_param_groups(self, keys):
        pgroups = {}
        for substr in keys:
            pgroups[substr] = {k: t for k, t in self.named_parameters() if substr in k}

        # Verify each param has at least one group
        unique_group_params = set([s for p in [list(pgroups[k].keys()) for k in keys]
                                       for s in p])
        assert len(unique_group_params) == len(list(self.named_parameters()))

        return pgroups

    @staticmethod
    def _allocate_out_tensors(x1_pyramid, batch_size):
        _, _, h_x1, w_x1, = x1_pyramid[0].size()
        flow_f = torch.zeros(batch_size, 2, h_x1, w_x1).to(dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device)
        flow_b = torch.zeros(batch_size, 2, h_x1, w_x1).to(dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device)
        occ_f = torch.zeros(batch_size, 1, h_x1, w_x1).to(dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device)
        occ_b = torch.zeros(batch_size, 1, h_x1, w_x1).to(dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device)
        return flow_f, flow_b, occ_f, occ_b

    def _warp(self, l, x1, x2, flow_f, flow_b, occ_f, occ_b, height_im, width_im):
        if l == 0:
            x2_warp = x2
            x1_warp = x1
        else:
            flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
            flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
            occ_f = upsample2d_as(occ_f, x1, mode="bilinear")
            occ_b = upsample2d_as(occ_b, x2, mode="bilinear")
            x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.args.div_flow)
            x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.args.div_flow)
        return x1_warp, x2_warp, flow_f, flow_b, occ_f, occ_b

    @staticmethod
    def collate_corr(corr, first_inp):
        b, ph, pw, h, w = corr.size()
        output_collated = corr.view(b, ph * pw, h, w)
        return output_collated / float(first_inp.size(1))

    def _correlate(self, x1, x2_warp):
        out_corr = self.correlation(x1, x2_warp)
        out_corr = out_corr.view(out_corr.shape[0], -1, out_corr.shape[3], out_corr.shape[4])
        out_corr = out_corr / x1.shape[1]
        # out_corr = self.collate_corr(out_corr, x1)
        out_corr_relu = self.leakyRELU(out_corr)
        return out_corr_relu

    def _squash(self, l, x1, x2):
        if l != self.output_level:
            x1_1by1 = self.conv_1x1[l](x1)
            x2_1by1 = self.conv_1x1[l](x2)
        else:
            x1_1by1 = x1
            x2_1by1 = x2
        return x1_1by1, x2_1by1

    def _rescale_flow(self, flow_f, flow_b, height_im, width_im):
        flow_f = rescale_flow(flow_f, self.args.div_flow, width_im, height_im, to_local=True)
        flow_b = rescale_flow(flow_b, self.args.div_flow, width_im, height_im, to_local=True)
        return flow_f, flow_b

    def _estimate_flow(self, out_corr_relu_f, out_corr_relu_b, x1_1by1, x2_1by1,
                       flow_f, flow_b):

        # Estimate forward and backward residual flows
        x_intm_f, flow_res_f = self.flow_estimators(
            torch.cat([out_corr_relu_f, x1_1by1, flow_f], dim=1))
        x_intm_b, flow_res_b = self.flow_estimators(
            torch.cat([out_corr_relu_b, x2_1by1, flow_b], dim=1))

        # Add residual flow to the aggregated flow
        flow_est_f = flow_f + flow_res_f
        flow_est_b = flow_b + flow_res_b

        # Estimate residual context based refinement
        ctxf = self.context_networks(torch.cat([x_intm_f, flow_est_f], dim=1))
        ctxb = self.context_networks(torch.cat([x_intm_b, flow_est_b], dim=1))
        flow_cont_f = flow_est_f + ctxf
        flow_cont_b = flow_est_b + ctxb

        return flow_cont_f, flow_cont_b

    def _estimate_occ(self, out_corr_relu_f, out_corr_relu_b,  x1_1by1, x2_1by1,
                      occ_f, occ_b):

        # Estimate residual occlusions
        x_intm_occ_f, occ_res_f = self.occ_estimators(
            torch.cat([out_corr_relu_f, x1_1by1, occ_f], dim=1))
        x_intm_occ_b, occ_res_b = self.occ_estimators(
            torch.cat([out_corr_relu_b, x2_1by1, occ_b], dim=1))

        # Add residual occlusions to the aggregated occlusions
        occ_est_f = occ_f + occ_res_f
        occ_est_b = occ_b + occ_res_b

        # Estimate residual context based refinement
        occ_cont_f = occ_est_f + self.occ_context_networks(
            torch.cat([x_intm_occ_f, occ_est_f], dim=1))
        occ_cont_b = occ_est_b + self.occ_context_networks(
            torch.cat([x_intm_occ_b, occ_est_b], dim=1))
        return occ_cont_f, occ_cont_b

    def _resize_and_warp_inputs(self, x1_raw, x2_raw, flow_f, flow_b,
                                flow_cont_f, flow_cont_b,
                                height_im, width_im):

        # Resizing input images for warping
        img1_resize = upsample2d_as(x1_raw, flow_f, mode="bilinear")
        img2_resize = upsample2d_as(x2_raw, flow_b, mode="bilinear")

        # Rescaling flow to the layer's size
        rescaled_f = rescale_flow(flow_cont_f, self.args.div_flow, width_im, height_im,
                                  to_local=False)
        rescaled_b = rescale_flow(flow_cont_b, self.args.div_flow, width_im, height_im,
                                  to_local=False)

        # Warping resized images
        img2_warp = self.warping_layer(img2_resize, rescaled_f, height_im, width_im,
                                       self.args.div_flow)
        img1_warp = self.warping_layer(img1_resize, rescaled_b, height_im, width_im,
                                       self.args.div_flow)

        return img1_resize, img2_resize, img1_warp, img2_warp, rescaled_f, rescaled_b

    def _refine_flow(self, flow_cont_f, flow_cont_b, img1_resize, img2_resize, img1_warp,
                     img2_warp, x1_1by1, x2_1by1, height_im, width_im):
        flow_f = self.refine_flow(flow_cont_f.detach(), img1_resize - img2_warp, x1_1by1)
        flow_b = self.refine_flow(flow_cont_b.detach(), img2_resize - img1_warp, x2_1by1)

        flow_cont_f = rescale_flow(flow_cont_f, self.args.div_flow, width_im, height_im,
                                   to_local=False)
        flow_cont_b = rescale_flow(flow_cont_b, self.args.div_flow, width_im, height_im,
                                   to_local=False)
        flow_f = rescale_flow(flow_f, self.args.div_flow, width_im, height_im, to_local=False)
        flow_b = rescale_flow(flow_b, self.args.div_flow, width_im, height_im, to_local=False)
        return flow_f, flow_cont_f, flow_b, flow_cont_b

    def _refine_occ(self, x1_1by1, x2_1by1, flow_f, flow_b, occ_cont_f, occ_cont_b,
                    height_im, width_im):
        x2_1by1_warp = self.warping_layer(x2_1by1, flow_f, height_im, width_im,
                                          self.args.div_flow)
        x1_1by1_warp = self.warping_layer(x1_1by1, flow_b, height_im, width_im,
                                          self.args.div_flow)

        occ_f = self.refine_occ(occ_cont_f.detach(), x1_1by1, x1_1by1 - x2_1by1_warp)
        occ_b = self.refine_occ(occ_cont_b.detach(), x2_1by1, x2_1by1 - x1_1by1_warp)
        return occ_f, occ_cont_f, occ_b, occ_cont_f

    def occ_upsampling(self, l, x1, x2, flow_f, flow_b, occ_f, occ_b,
                       height_im, width_im):
        x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.args.div_flow)
        x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.args.div_flow)
        flow_b_warp = self.warping_layer(flow_b, flow_f, height_im, width_im,
                                         self.args.div_flow)
        flow_f_warp = self.warping_layer(flow_f, flow_b, height_im, width_im,
                                         self.args.div_flow)

        if l != self.num_levels - 1:
            x1_in = self.conv_1x1_1(x1)
            x2_in = self.conv_1x1_1(x2)
            x1_w_in = self.conv_1x1_1(x1_warp)
            x2_w_in = self.conv_1x1_1(x2_warp)
        else:
            x1_in = x1
            x2_in = x2
            x1_w_in = x1_warp
            x2_w_in = x2_warp

        occ_f = self.occ_shuffle_upsample(occ_f,
                                          torch.cat([x1_in, x2_w_in, flow_f, flow_b_warp],
                                                    dim=1))
        occ_b = self.occ_shuffle_upsample(occ_b,
                                          torch.cat([x2_in, x1_w_in, flow_b, flow_f_warp],
                                                    dim=1))
        return occ_f, occ_b



import math
import torch
from torch import nn
import torch.nn.functional as F

__all__ = [
    'flow_warp', 'vector2density', 'density2vector', 'prob_gather',
    'disp2flow', 'downsample_flow', 'resize_dense_vector'
]


def flow_warp(x, flo, mul=True):
    """
    inverse warp an image/tensor (im2) back to im1, according to the optical flow
    x: [B, C, H, W] (im2)
    flo: [B, 2, H, W] flow
    """
    B, C, H, W = x.size()
    # mesh grid
    xx = torch.arange(0, W).view(1, -1).repeat(H, 1).to(x.device)
    yy = torch.arange(0, H).view(-1, 1).repeat(1, W).to(x.device)
    xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
    yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
    grid = torch.cat((xx, yy), 1).float()

    vgrid = grid + flo

    # scale grid to [-1,1]
    vgrid = torch.stack([
        2.0 * vgrid[:, 0, :, :] / max(W - 1, 1) - 1.0,
        2.0 * vgrid[:, 1, :, :] / max(H - 1, 1) - 1.0
    ],
                        dim=1)

    vgrid = vgrid.permute(0, 2, 3, 1)
    output = F.grid_sample(x, vgrid, padding_mode='border', align_corners=True)
    mask = torch.ones(x.size(), device=x.device)
    mask = F.grid_sample(mask, vgrid, padding_mode='zeros', align_corners=True)

    mask[mask < 0.9999] = 0
    mask[mask > 0] = 1

    if mul:
        return output * mask
    else:
        return output, mask


def vector2density(vect, c, dim):
    # vect: point estimate
    # c: support size
    # dim: dimension (stereo: 1 / flow: 2)
    assert vect.size(1) == dim
    if dim == 2:
        return _flow2distribution(vect, c)
    else:
        dist = _flow2distribution(disp2flow(vect), c)
        return dist[:, c * (2 * c + 1):(c + 1) * (2 * c + 1), :, :]


def density2vector(prob, dim, normalize=True):
    # prob: density
    # dim: dimension (stereo: 1 / flow: 2)
    # normalize: if true, prob will be normalized via softmax
    if dim == 2:
        flow = _prob2flow(prob, normalize)
        return flow
    else:
        prob_padded = _disp_prob2flow_prob(prob, normalize)
        flow = _prob2flow(prob_padded, False)
        disp = flow[:, 0, :, :].unsqueeze(1)
        return disp


def prob_gather(prob, normalize=True, dim=2, return_indices=False):
    # gather probability for confidence map visualization
    # return shape: out, indice [B,1,H,W]
    if normalize:
        prob = F.softmax(prob, dim=1)
    if dim == 1:
        prob = _disp_prob2flow_prob(prob, False)
    B, C, H, W = prob.size()
    d = int(math.sqrt(C))
    pr = prob.reshape(B, d, d, -1).permute(0, 3, 1, 2)
    avg_pool = nn.AvgPool2d(kernel_size=2, stride=1, padding=0)
    max_pool = nn.MaxPool2d(kernel_size=d - 1, stride=1, return_indices=True)
    out, indice = max_pool(avg_pool(pr))
    out = 4 * out.squeeze().reshape(B, 1, H, W)
    if not return_indices:
        return out
    else:
        indice += indice / (d - 1)
        indice = indice.squeeze().reshape(B, 1, H, W)
        return out, indice


def disp2flow(disp):
    assert disp.size(1) == 1
    padder = torch.zeros(disp.size(), device=disp.device)
    return torch.cat([disp, padder], dim=1)


def downsample_flow(flo, scale_factor):
    # note the value is not scaled accordingly here
    # return downsampled flow field and valid mask
    assert scale_factor <= 1
    B, C, H, W = flo.size()
    if flo.size(1) == 2:
        # dense format
        flo = F.interpolate(
            flo,
            scale_factor=scale_factor,
            mode='bilinear',
            align_corners=True)
        mask = torch.ones((B, 1, int(H * scale_factor), int(W * scale_factor)),
                          dtype=torch.float,
                          device=flo.device)
    else:
        # sparse format
        flo = F.avg_pool2d(flo, int(1 / scale_factor))
        mask = (flo[:, 2, :, :].unsqueeze(1) > 0).float()
        flo = flo[:, :2, :, :] / (flo[:, 2, :, :].unsqueeze(1) + 1e-9)
    return flo, mask


def resize_dense_vector(vec, des_height, des_width):
    ratio_height = float(des_height / vec.size(2))
    ratio_width = float(des_width / vec.size(3))
    vec = F.interpolate(
        vec, (des_height, des_width), mode='bilinear', align_corners=True)
    if vec.size(1) == 1:
        vec = vec * ratio_width
    else:
        vec = torch.stack(
            [vec[:, 0, :, :] * ratio_width, vec[:, 1, :, :] * ratio_height],
            dim=1)
    return vec


def _flow2distribution(flo, c):
    # flo: [B,2,H,W]
    # return [B,d*d,H,W], d=2c+1
    B, _, H, W = flo.size()
    flo = torch.clamp(flo, min=-c, max=c)
    x = flo[:, 0, :, :]
    y = flo[:, 1, :, :]
    x0 = torch.floor(x)
    x1 = x0 + 1
    y0 = torch.floor(y)
    y1 = y0 + 1

    x0_safe = torch.clamp(x0, min=-c, max=c)
    y0_safe = torch.clamp(y0, min=-c, max=c)
    x1_safe = torch.clamp(x1, min=-c, max=c)
    y1_safe = torch.clamp(y1, min=-c, max=c)

    wt_x0 = (x1 - x) * torch.eq(x0, x0_safe).float()
    wt_x1 = (x - x0) * torch.eq(x1, x1_safe).float()
    wt_y0 = (y1 - y) * torch.eq(y0, y0_safe).float()
    wt_y1 = (y - y0) * torch.eq(y1, y1_safe).float()

    wt_tl = wt_x0 * wt_y0
    wt_tr = wt_x1 * wt_y0
    wt_bl = wt_x0 * wt_y1
    wt_br = wt_x1 * wt_y1

    mask_tl = torch.eq(x0, x0_safe).float() * torch.eq(y0, y0_safe).float()
    mask_tr = torch.eq(x1, x1_safe).float() * torch.eq(y0, y0_safe).float()
    mask_bl = torch.eq(x0, x0_safe).float() * torch.eq(y1, y1_safe).float()
    mask_br = torch.eq(x1, x1_safe).float() * torch.eq(y1, y1_safe).float()

    wt_tl *= mask_tl
    wt_tr *= mask_tr
    wt_bl *= mask_bl
    wt_br *= mask_br

    out = torch.zeros((B, (2 * c + 1)**2, H, W), device=flo.device)
    label_tl = (y0_safe + c) * (2 * c + 1) + x0_safe + c
    label_tr = (y0_safe + c) * (2 * c + 1) + x1_safe + c
    label_bl = (y1_safe + c) * (2 * c + 1) + x0_safe + c
    label_br = (y1_safe + c) * (2 * c + 1) + x1_safe + c

    out.scatter_add_(1, label_tl.unsqueeze(1).long(), wt_tl.unsqueeze(1))
    out.scatter_add_(1, label_tr.unsqueeze(1).long(), wt_tr.unsqueeze(1))
    out.scatter_add_(1, label_bl.unsqueeze(1).long(), wt_bl.unsqueeze(1))
    out.scatter_add_(1, label_br.unsqueeze(1).long(), wt_br.unsqueeze(1))

    return out


def _prob2cornerflow(prob, normalize=True):
    # prob: [B,C,H,W]
    def indice2flow(ind, d):
        # ind [B,1,H,W]
        return torch.cat([ind % d - d // 2, ind // d - d // 2], 1)

    if normalize:
        normalizer = nn.Softmax(dim=1)
        prob = normalizer(prob)
    B, C, H, W = prob.size()
    d = int(math.sqrt(C))
    pr = prob.reshape(B, d, d, -1).permute(0, 3, 1, 2)
    avg_pool = nn.AvgPool2d(kernel_size=2, stride=1, padding=0)
    max_pool = nn.MaxPool2d(kernel_size=d - 1, stride=1, return_indices=True)
    out, indice = max_pool(avg_pool(pr))
    indice += indice // (d - 1)  # in original coordinate
    indice = indice.squeeze().reshape(B, H, W).unsqueeze(1)
    lt_prob = torch.gather(prob, 1, indice)
    lt_flow = indice2flow(indice, d).float()
    rt_prob = torch.gather(prob, 1, indice + 1)
    rt_flow = indice2flow(indice + 1, d).float()
    lb_prob = torch.gather(prob, 1, indice + d)
    lb_flow = indice2flow(indice + d, d).float()
    rb_prob = torch.gather(prob, 1, indice + d + 1)
    rb_flow = indice2flow(indice + d + 1, d).float()
    return [lt_prob, rt_prob, lb_prob,
            rb_prob], [lt_flow, rt_flow, lb_flow, rb_flow]


def _cornerflow2expectation(cor_prob, cor_flow):
    cor_prob_sum = sum(cor_prob)
    cor_prob_n = [prob / cor_prob_sum for prob in cor_prob]
    out = torch.cat([
        cor_flow[1][:, 0, :, :].unsqueeze(1) - cor_prob_n[0] - cor_prob_n[2],
        cor_flow[2][:, 1, :, :].unsqueeze(1) - cor_prob_n[0] - cor_prob_n[1]
    ], 1)
    return out


def _prob2flow(prob, normalize=True):
    cor_prob, cor_flow = _prob2cornerflow(prob, normalize)
    out = _cornerflow2expectation(cor_prob, cor_flow)
    return out


def _disp_prob2flow_prob(prob, normalize=True):
    ###### return NORMALIZED FLOW PROBABILITY ######
    if normalize:
        normalizer = nn.Softmax(dim=1)
        prob = normalizer(prob)
    B, d, H, W = prob.size()
    padding = torch.zeros((B, d * (d - 1) // 2, H, W), device=prob.device)
    prob_padded = torch.cat([padding, prob, padding], dim=1)
    return prob_padded



"""Abstract model to be used as a parent of concrete optical flow networks.

In order to use this module, the concrete model should just need to:
1. Specify the network structure in their __init__() method and call super().__init__() with the required args.
2. Implement a forward method which receives inputs and outputs as specified. See the forward method docs for more details.
3. Optionally, define a loss function, if the model is able to be trained.
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
import warnings
from abc import abstractmethod
from argparse import ArgumentParser, Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

with warnings.catch_warnings():
    warnings.simplefilter('ignore')  # Workaround until pl stop raising the metrics deprecation warning
    import pytorch_lightning as pl
import torch
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset

from ptlflow.data import flow_transforms as ft
from ptlflow.data.datasets import (
    FlyingChairsDataset, FlyingChairs2Dataset, Hd1kDataset, KittiDataset, SintelDataset, FlyingThings3DDataset,
    FlyingThings3DSubsetDataset)
from ptlflow.utils.external.raft import InputPadder
from ptlflow.utils.utils import config_logging, make_divisible
from ptlflow.utils.flow_metrics import FlowMetrics

config_logging()


class BaseModel(pl.LightningModule):
    """A base abstract optical flow model."""

    def __init__(
        self,
        args: Namespace,
        loss_fn: Callable,
        output_stride: int
    ) -> None:
        """Initialize BaseModel.

        Parameters
        ----------
        args : Namespace
            A namespace with the required arguments. Typically, this can be gotten from add_model_specific_args().
        loss_fn : Callable
            A function to be used to compute the loss for the training. The input of this function must match the output of the
            forward() method. The output of this function must be a tensor with a single value.
        output_stride : int
            How many times the output of the network is smaller than the input.
        """
        super(BaseModel, self).__init__()

        self.args = args
        self.loss_fn = loss_fn
        self.output_stride = output_stride

        self.train_metrics = FlowMetrics(prefix='train/')
        self.val_metrics = FlowMetrics(prefix='val/')

        self.train_dataloader_length = 0
        self.train_epoch_step = 0

        self.val_dataloader_names = []
        self.val_dataloader_lengths = []

        self.last_inputs = None
        self.last_predictions = None

        self.save_hyperparameters()

    @staticmethod
    def add_model_specific_args(
        parent_parser: Optional[ArgumentParser] = None,
    ) -> ArgumentParser:
        """Generate a parser for the arguments required by this model.

        Parameters
        ----------
        parent_parser : ArgumentParser
            An existing parser, to be extended with the arguments from this model.

        Returns
        -------
        ArgumentParser
            The parser after extending its arguments.

        Notes
        -----
        If the concrete model needs to add more arguments than these defined in this BaseModel, then it should create its
        own method defined as follows:

        >>>
        @staticmethod
        def add_model_specific_args(parent_parser):
            parent_parser = BaseModel.add_model_specific_args(parent_parser)
            parser = ArgumentParser(parents=[parent_parser], add_help=False)
            parser.add_argument(...add new arguments...)
            return parser
        """
        parents = [parent_parser] if parent_parser is not None else []
        add_help = False if parent_parser is not None else True
        parser = ArgumentParser(parents=parents, add_help=add_help)
        parser.add_argument('--train_batch_size', type=int, default=0, help='')
        parser.add_argument('--train_num_workers', type=int, default=4, help='')
        parser.add_argument('--train_transform_cuda', action='store_true', default=False, help='')
        parser.add_argument('--train_transform_fp16', action='store_true', default=False, help='')
        parser.add_argument('--lr', type=float, default=1e-4)
        parser.add_argument('--wdecay', type=float, default=1e-4)
        parser.add_argument(
            '--train_dataset', type=str, default=None,
            help=('String specifying the dataset for training. See the docs of '
                  'ptlflow.models.base_model.base_model.BaseModel.parse_dataset_selection for more details about this '
                  'string.'))
        parser.add_argument('--train_crop_size', type=int, nargs=2, default=None, help='')
        parser.add_argument(
            '--val_dataset', type=str, default=None,
            help=('String specifying the dataset for validation. See the docs of '
                  'ptlflow.models.base_model.base_model.BaseModel.parse_dataset_selection for more details about this '
                  'string.'))
        parser.add_argument(
            '--pretrained_ckpt', type=str, default=None,
            help=('A string identifier of the pretrained checkpoint to load. The string will be first checked against the '
                  'key names in the model pretrained_checkpoints class attribute. If it is not found, then it will be treated '
                  'as a path to a local file.'))
        return parser

    @abstractmethod
    def forward(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        """Forward the inputs through the network and produce the predictions.

        The method inputs can be anything, up to the implementation of the concrete model. However, the recommended input is
        to receive only dict[str, Any] as argument. This dict should contain everything required for one pass of the network
        (images, etc.). Arguments which do not change for each forward should be defined as arguments in the parser
        (see add_model_specific_args()).

        Parameters
        ----------
        args : Any
            Any arguments.
        kwargs : Any
            Any named arguments.

        Returns
        -------
        Dict[str, torch.Tensor]
            For compatibility with the framework, the output should be a dict containing at least the following keys:

            - 'flows': a 5D tensor BNCHW containing the predicted flows. The flows must be at the original scale, i.e.,
              their size is the same as the input images, and the flow magnitudes are scaled accordingly. Most networks only
              produce a single 2D optical flow per batch element, so the output shape will be B12HW. N can be larger than one
              if the network produces predictions for a larger temporal window.

            - 'occs': optional, and only included if the network also predicts occlusion masks. It is a 5D tensor following the
              same structure as 'flows'.

            - 'mbs': same as 'occs' but for occlusion masks.

            - 'confs': same as 'occs' but for flow confidence predictions.


            The keys above are used by other parts of PTLFlow (if available). The output can also include any other keys as
            well. For example, by default, the output of forward() will be passed to the loss function. So the output may
            include keys which are going to be used for computing the training loss.
        """
        pass

    def training_step(
        self,
        batch: Dict[str, Any],
        batch_idx: int
    ) -> Dict[str, Any]:
        """Perform one step of the training.

        This function is called internally by Pytorch-Lightning during training.

        Parameters
        ----------
        batch : Dict[str, Any]
            One batch of data, that is going to be given as input to the network.
        batch_idx : int
            The index of the current batch.

        Returns
        -------
        Dict[str, Any]
            A dict with two keys:

            - 'loss': torch.Tensor, containing the loss value. Required by Pytorch-Lightning for the optimization step.

            - 'dataset_name': str, a string representing the name of the dataset from where this batch came from. Used only for
              logging purposes.
        """
        preds = self(batch)
        self.last_inputs = batch
        self.last_predictions = preds
        loss = self.loss_fn(preds, batch)
        if isinstance(loss, dict):
            loss = loss['loss']
        metrics = self.train_metrics(preds, batch)
        metrics['train/loss'] = loss.item()
        self.log_dict(metrics, on_step=True, on_epoch=True)
        self.log('epe', metrics['train/epe'], prog_bar=True, on_step=True, on_epoch=True)

        outputs = {'loss': loss,
                   'dataset_name': batch['meta']['dataset_name']}
        return outputs

    def validation_step(
        self,
        batch: Dict[str, Any],
        batch_idx: int,
        dataloader_idx: int = 0
    ) -> Dict[str, Any]:
        """Perform one step of the validation.

        This function is called internally by Pytorch-Lightning during validation.

        Parameters
        ----------
        batch : Dict[str, Any]
            One batch of data, that is going to be given as input to the network.
        batch_idx : int
            The index of the current batch.
        dataloader_idx : int, default 0
            When using multiple loaders, indicate from which loader this input is coming from.

        Returns
        -------
        Dict[str, Any]
            A dict with two keys:

            - 'metrics': dict[str, torch.Tensor], the metrics computed during this validation step.

            - 'dataset_name': str, a string representing the name of the dataset from where this batch came from. Used only for
              logging purposes.

        See Also
        --------
        ptlflow.utils.flow_metrics.FlowMetrics : class to manage and compute the optical flow metrics.
        """
        padder = InputPadder(batch['images'].shape, stride=self.output_stride)
        batch['images'] = padder.pad(batch['images'])
        preds = self(batch)
        batch['images'] = padder.unpad(batch['images'])
        keys = ['flows', 'occs', 'mbs', 'confs']
        for k in keys:
            if k in preds:
                preds[k] = padder.unpad(preds[k])
        self.last_inputs = batch
        self.last_predictions = preds
        metrics = self.val_metrics(preds, batch)
        train_val_metrics = self._split_train_val_metrics(metrics, batch.get('meta'))

        outputs = {'metrics': train_val_metrics,
                   'dataset_name': batch['meta']['dataset_name']}
        return outputs

    def validation_epoch_end(
        self,
        outputs: Union[List[Dict[str, torch.Tensor]], List[List[Dict[str, torch.Tensor]]]]
    ) -> None:
        """Perform operations at the end of one validation epoch.

        This function is called internally by Pytorch-Lightning during validation.

        Parameters
        ----------
        outputs : Union[List[Dict[str, Any]], List[List[Dict[str, Any]]]]
            A list in which each element is the output of validation_step().

        See Also
        --------
        validation_step
        """
        if not isinstance(outputs[0], list):
            outputs = [outputs]

        for i, loader_outputs in enumerate(outputs):
            # Log mean metrics for one val dataloader
            metrics_cum = {}
            for output_dict in loader_outputs:
                for k, v in output_dict['metrics'].items():
                    if metrics_cum.get(k) is None:
                        metrics_cum[k] = [0.0, 0]
                    metrics_cum[k][0] += v
                    metrics_cum[k][1] += 1
            metrics_mean = {}
            for k, v in metrics_cum.items():
                metrics_mean[k] = v[0] / v[1]
            self.log_dict(metrics_mean)

            # Find the EPE metric on the full split
            epe_key = [k for k in metrics_mean if 'full' in k and 'epe' in k][0]
            self.log(self.val_dataloader_names[i], metrics_mean[epe_key], prog_bar=True)

    def configure_optimizers(self) -> Dict[str, Any]:
        """Initialize the optimizers and LR schedulers.

        This function is called internally by Pytorch-Lightning at the beginning of the training.

        Returns
        -------
        Dict[str, Any]
            A dict with two keys:
            - 'optimizer': an optimizer from PyTorch.
            - 'lr_scheduler': Dict['str', Any], a dict with the selected scheduler and its required arguments.
        """
        assert self.loss_fn is not None, f'Model {self.__class__.__name__} cannot be trained. It does not have loss function.'

        self.train_dataloader()  # Just to initialize dataloader variables

        if self.args.max_steps is None:
            if self.args.max_epochs is None:
                self.args.max_epochs = 10
                logging.warning('--max_epochs is not set. It will be set to %d.', self.args.max_epochs)

            gpu_divider = self.args.gpus
            if isinstance(gpu_divider, list) or isinstance(gpu_divider, tuple):
                gpu_divider = len(gpu_divider)
            elif isinstance(gpu_divider, str):
                gpu_divider = len(gpu_divider.split(','))
            elif not isinstance(gpu_divider, int):
                gpu_divider = 1

            self.args.max_steps = self.args.max_epochs * (self.train_dataloader_length // gpu_divider)

        optimizer = optim.AdamW(self.parameters(), lr=self.args.lr, weight_decay=self.args.wdecay)
        lr_scheduler = optim.lr_scheduler.OneCycleLR(
            optimizer, self.args.lr, total_steps=self.args.max_steps, pct_start=0.05, cycle_momentum=False, anneal_strategy='linear')
        return {'optimizer': optimizer,
                'lr_scheduler': {'scheduler': lr_scheduler, 'interval': 'step'}}

    def train_dataloader(self) -> DataLoader:
        """Initialize and return the training dataloader.

        self.args.train_dataset will be parsed into the selected datasets and their parameters. parse_dataset_selection() is
        used to parse, and the parsed outputs are used as follows: (dataset_multiplier, dataset_name, dataset_params...).

        A method called _get_[dataset_name]_dataset(dataset_params) will be called to get the instance of each dataset.

        Each dataset will be repeated dataset_multiplier times and then mixed together into a single dataloader with samples
        of all the datasets.

        Returns
        -------
        DataLoader
            A single dataloader with all the selected training datasets.

        Notes
        -----
        This function could be easily modified to return one dataloader for each dataset, which may be preferable in some
        cases. However, this would require that the loading function have some logic to mix all the dataloaders internally.
        In order to keep the whole BaseModel more generic, we use just a single dataloader randomly mixed to train.

        See Also
        --------
        parse_dataset_selection
            This function does the parsing of self.args.train_dataset. You can read its documentation to learn how to write
            valid dataset strings.
        val_dataloader
            Similar to this method, but val_dataloader returns one dataloader for each dataset.
        """
        if self.args.train_dataset is None:
            self.args.train_dataset = 'chairs-train'
            logging.warning('--train_dataset is not set. It will be set as %s', self.args.train_dataset)
        if self.args.train_batch_size == 0:
            self.args.train_batch_size = 8
            logging.warning('--train_batch_size is not set. It will be set to %d', self.args.train_batch_size)

        if self.args.train_dataset is not None:
            parsed_datasets = self.parse_dataset_selection(self.args.train_dataset)
            train_dataset = None
            for parsed_vals in parsed_datasets:
                multiplier = parsed_vals[0]
                dataset_name = parsed_vals[1]
                dataset = getattr(self, f'_get_{dataset_name}_dataset')(True, *parsed_vals[2:])
                for _ in range(multiplier):
                    if train_dataset is None:
                        train_dataset = dataset
                    else:
                        train_dataset += dataset

            train_pin_memory = False if self.args.train_transform_cuda else True
            train_dataloader = DataLoader(
                train_dataset, self.args.train_batch_size, shuffle=True, num_workers=self.args.train_num_workers,
                pin_memory=train_pin_memory, drop_last=False)
            self.train_dataloader_length = len(train_dataloader)
            return train_dataloader

    def val_dataloader(self) -> Optional[List[DataLoader]]:
        """Initialize and return the list of validation dataloaders.

        self.args.val_dataset will be parsed into the selected datasets and their parameters. parse_dataset_selection() is
        used to parse, and the parsed outputs are used as follows: (dataset_multiplier, dataset_name, dataset_params...).

        A method called _get_[dataset_name]_dataset(dataset_params) will be called to get the instance of each dataset.

        If there is a dataset_multiplier, it will be ignored. Each dataset will be returned in a separated dataloader.

        Returns
        -------
        Optional[List[DataLoader]]
            A list of dataloaders each for one dataset.

        See Also
        --------
        parse_dataset_selection
            This function does the parsing of self.args.val_dataset. You can read its documentation to learn how to write
            valid dataset strings.
        """
        if self.args.val_dataset is None:
            self.args.val_dataset = 'sintel-final-trainval+sintel-clean-trainval+kitti-2012-trainval+kitti-2015-trainval'
            logging.warning(
                '--val_dataset is not set. It will be set as %s. If you want to skip validation, then set --val_dataset none.',
                self.args.val_dataset)
        elif self.args.val_dataset.lower() == 'none':
            return None

        parsed_datasets = self.parse_dataset_selection(self.args.val_dataset)
        dataloaders = []
        self.val_dataloader_names = []
        self.val_dataloader_lengths = []
        for parsed_vals in parsed_datasets:
            dataset_name = parsed_vals[1]
            dataset = getattr(self, f'_get_{dataset_name}_dataset')(False, *parsed_vals[2:])
            dataloaders.append(DataLoader(dataset, 1, shuffle=False, num_workers=1, pin_memory=False, drop_last=False))

            self.val_dataloader_names.append('-'.join(parsed_vals[1:]))
            self.val_dataloader_lengths.append(len(dataset))

        return dataloaders

    def parse_dataset_selection(
        self,
        dataset_selection: str,
    ) -> List[Tuple[str, int]]:
        """Parse the input string into the selected dataset and their multipliers and parameters.

        For example, 'chairs-train+3*sintel-clean-trainval+kitti-2012-train*5' will be parsed into
        [(1, 'chairs', 'train'), (3, 'sintel', 'clean', 'trainval'), (5, 'kitti', '2012', 'train')].

        Parameters
        ----------
        dataset_selection : str
            The string defining the dataset selection. Each dataset is separated by a '+' sign. The multiplier must be either
            in the beginning or the end of one dataset string, connected to a '*' sign. The remaining content must be separated
            by '-' symbols.

        Returns
        -------
        List[Tuple[str, int]]
            The parsed choice of datasets and their number of repetitions.

        Raises
        ------
        ValueError
            If the given string is invalid.
        """
        if dataset_selection is None:
            return []

        dataset_selection = dataset_selection.replace(' ', '')
        datasets = dataset_selection.split('+')
        for i in range(len(datasets)):
            tokens = datasets[i].split('*')
            if len(tokens) == 1:
                datasets[i] = (1,) + tuple(tokens[0].split('-'))
            elif len(tokens) == 2:
                try:
                    mult, params = int(tokens[0]), tokens[1]
                except ValueError:
                    params, mult = tokens[0], int(tokens[1])  # if the multiplier comes last.
                datasets[i] = (mult,) + tuple(params.split('-'))
            else:
                raise ValueError(
                    'The specified dataset string {:} is invalid. Check the BaseModel.parse_dataset_selection() documentation '
                    'to see how to write a valid selection string.')
        return datasets

    ###########################################################################
    # Logging
    ###########################################################################

    def _split_train_val_metrics(
        self,
        metrics: Dict[str, float],
        inputs_meta: Optional[Dict[str, Any]] = None
    ) -> Dict[str, float]:
        """Split the val metrics into 'train', 'val', and 'full' categories.

        This is useful for producing metrics for different subsets without having to forward the same sample multiple times.
        When computing metrics for the full validation dataset, the metrics for the 'train' and 'val' subsets are also
        computed at the same time.

        Parameters
        ----------
        metrics : dict[str, float]
            The metrics computed after one validation step.
        inputs_meta : Optional[Dict[str, Any]], optional
            If this is not provided, then the metrics are not split. This is a dict with two keys:
            - 'dataset_name': str, a string representing the name of the dataset these metrics correspond to.
            - 'is_val': bool, indicating whether these metrics come from a sample in the 'val' split or not.

        Returns
        -------
        dict[str, float]
            The metrics after being split.

        See Also
        --------
        validation_step
        """
        dataset_name = None
        if inputs_meta is not None and inputs_meta.get('dataset_name') is not None:
            dataset_name = inputs_meta['dataset_name'][0]

        log_metrics = {}
        for k, v in metrics.items():
            if dataset_name is not None:
                log_metrics[f'val_{dataset_name}/full/{k}'] = v
            else:
                log_metrics[f'val/full/{k}'] = v

            if inputs_meta is not None and inputs_meta.get('is_val') is not None:
                if inputs_meta['is_val'][0]:
                    split = 'val'
                else:
                    split = 'train'

                if dataset_name is not None:
                    log_metrics[f'val_{dataset_name}/{split}/{k}'] = v
                else:
                    log_metrics[f'val/{split}/{k}'] = v

        return log_metrics

    ###########################################################################
    # _get_datasets
    ###########################################################################

    def _get_chairs_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        device = 'cuda' if self.args.train_transform_cuda else 'cpu'
        md = make_divisible

        if is_train:
            if self.args.train_crop_size is None:
                cy, cx = (md(368, self.output_stride), md(496, self.output_stride))
                self.args.train_crop_size = (cy, cx)
                logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
            else:
                cy, cx = (
                    md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose([
                ft.ToTensor(device=device, fp16=self.args.train_transform_fp16),
                ft.RandomScaleAndCrop((cy, cx), (-0.1, 1.0), (-0.2, 0.2), min_pool_binary=True),
                ft.ColorJitter(0.4, 0.4, 0.4, 0.5/3.14, 0.2),
                ft.GaussianNoise(0.02),
                ft.RandomPatchEraser(0.5, (int(1), int(3)), (int(50), int(100)), 'mean'),
                ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
            ])
        else:
            transform = ft.ToTensor()

        split = 'trainval'
        if len(args) > 0 and args[0] in ['train', 'val', 'trainval']:
            split = args[0]
        dataset = FlyingChairsDataset(self.args.flying_chairs_root_dir, split=split, transform=transform)
        return dataset

    def _get_chairs2_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        device = 'cuda' if self.args.train_transform_cuda else 'cpu'
        md = make_divisible

        if is_train:
            if self.args.train_crop_size is None:
                cy, cx = (md(368, self.output_stride), md(496, self.output_stride))
                self.args.train_crop_size = (cy, cx)
                logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
            else:
                cy, cx = (
                    md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose([
                ft.ToTensor(device=device, fp16=self.args.train_transform_fp16),
                ft.RandomScaleAndCrop((cy, cx), (-0.1, 1.0), (-0.2, 0.2), min_pool_binary=True),
                ft.ColorJitter(0.4, 0.4, 0.4, 0.5/3.14, 0.2),
                ft.GaussianNoise(0.02),
                ft.RandomPatchEraser(0.5, (int(1), int(3)), (int(50), int(100)), 'mean'),
                ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
            ])
        else:
            transform = ft.ToTensor()

        split = 'trainval'
        if len(args) > 0 and args[0] in ['train', 'val', 'trainval']:
            split = args[0]
        dataset = FlyingChairs2Dataset(
            self.args.flying_chairs2_root_dir, split=split, transform=transform, add_reverse=True, get_occlusion_mask=False,
            get_motion_boundary_mask=False, get_backward=False)
        return dataset

    def _get_hd1k_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        device = 'cuda' if self.args.train_transform_cuda else 'cpu'
        md = make_divisible

        if is_train:
            if self.args.train_crop_size is None:
                cy, cx = (md(368, self.output_stride), md(768, self.output_stride))
                self.args.train_crop_size = (cy, cx)
                logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
            else:
                cy, cx = (
                    md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose([
                ft.ToTensor(device=device, fp16=self.args.train_transform_fp16),
                ft.RandomScaleAndCrop((cy, cx), (-0.5, 0.2), (-0.2, 0.2), min_pool_binary=True),
                ft.ColorJitter(0.4, 0.4, 0.4, 0.5/3.14, 0.2),
                ft.GaussianNoise(0.02),
                ft.RandomPatchEraser(0.5, (int(1), int(3)), (int(50), int(100)), 'mean'),
                ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
            ])
        else:
            transform = ft.ToTensor()

        split = 'trainval'
        if len(args) > 0 and args[0] in ['train', 'val', 'trainval', 'test']:
            split = args[0]

        dataset = Hd1kDataset(self.args.hd1k_root_dir, split=split, transform=transform)
        return dataset

    def _get_kitti_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        device = 'cuda' if self.args.train_transform_cuda else 'cpu'
        md = make_divisible

        if is_train:
            if self.args.train_crop_size is None:
                cy, cx = (md(288, self.output_stride), md(960, self.output_stride))
                self.args.train_crop_size = (cy, cx)
                logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
            else:
                cy, cx = (
                    md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose([
                ft.ToTensor(device=device, fp16=self.args.train_transform_fp16),
                ft.RandomScaleAndCrop((cy, cx), (-0.2, 0.4), (-0.2, 0.2), min_pool_binary=True),
                ft.ColorJitter(0.4, 0.4, 0.4, 0.5/3.14, 0.2),
                ft.GaussianNoise(0.02),
                ft.RandomPatchEraser(0.5, (int(1), int(3)), (int(50), int(100)), 'mean'),
            ])
        else:
            transform = ft.ToTensor()

        versions = ['2012', '2015']
        split = 'trainval'
        if len(args) > 0:
            for v in args:
                if v in ['2012', '2015']:
                    versions = [v]
                elif v in ['train', 'val', 'trainval', 'test']:
                    split = v

        dataset = KittiDataset(
            self.args.kitti_2012_root_dir, self.args.kitti_2015_root_dir, versions=versions, split=split, transform=transform)
        return dataset

    def _get_sintel_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        device = 'cuda' if self.args.train_transform_cuda else 'cpu'
        md = make_divisible

        if is_train:
            if self.args.train_crop_size is None:
                cy, cx = (md(368, self.output_stride), md(768, self.output_stride))
                self.args.train_crop_size = (cy, cx)
                logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
            else:
                cy, cx = (
                    md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose([
                ft.ToTensor(device=device, fp16=self.args.train_transform_fp16),
                ft.RandomScaleAndCrop((cy, cx), (-0.2, 0.6), (-0.2, 0.2), min_pool_binary=True),
                ft.ColorJitter(0.4, 0.4, 0.4, 0.5/3.14, 0.2),
                ft.GaussianNoise(0.02),
                ft.RandomPatchEraser(0.5, (int(1), int(3)), (int(50), int(100)), 'mean'),
                ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
            ])
        else:
            transform = ft.ToTensor()

        pass_names = ['clean', 'final']
        split = 'trainval'
        if len(args) > 0:
            for v in args:
                if v in ['clean', 'final']:
                    pass_names = [v]
                elif v in ['train', 'val', 'trainval', 'test']:
                    split = v

        dataset = SintelDataset(
            self.args.mpi_sintel_root_dir, split=split, pass_names=pass_names, transform=transform,
            get_occlusion_mask=False)
        return dataset

    def _get_things_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        device = 'cuda' if self.args.train_transform_cuda else 'cpu'
        md = make_divisible

        if is_train:
            if self.args.train_crop_size is None:
                cy, cx = (md(400, self.output_stride), md(720, self.output_stride))
                self.args.train_crop_size = (cy, cx)
                logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
            else:
                cy, cx = (
                    md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose([
                ft.ToTensor(device=device, fp16=self.args.train_transform_fp16),
                ft.RandomScaleAndCrop((cy, cx), (-0.4, 0.8), (-0.2, 0.2), min_pool_binary=True),
                ft.ColorJitter(0.4, 0.4, 0.4, 0.5/3.14, 0.2),
                ft.GaussianNoise(0.02),
                ft.RandomPatchEraser(0.5, (int(1), int(3)), (int(50), int(100)), 'mean'),
                ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
            ])
        else:
            transform = ft.ToTensor()

        pass_names = ['clean', 'final']
        split = 'trainval'
        is_subset = False
        if len(args) > 0:
            for v in args:
                if v in ['clean', 'final']:
                    pass_names = [v]
                elif v in ['train', 'val', 'trainval']:
                    split = v
                elif v == 'subset':
                    is_subset = True

        if is_subset:
            dataset = FlyingThings3DSubsetDataset(
                self.args.flying_things3d_subset_root_dir, split=split, pass_names=pass_names,
                side_names=['left', 'right'], add_reverse=True, transform=transform,
                get_occlusion_mask=False, get_motion_boundary_mask=False, get_backward=False)
        else:
            dataset = FlyingThings3DDataset(
                self.args.flying_things3d_root_dir, split=split, pass_names=pass_names,
                side_names=['left', 'right'], add_reverse=True, transform=transform,
                get_occlusion_mask=False, get_motion_boundary_mask=False, get_backward=False)
        return dataset

    def _get_overfit_dataset(
        self,
        is_train: bool,
        *args: str
    ) -> Dataset:
        md = make_divisible
        if self.args.train_crop_size is None:
            cy, cx = (md(436, self.output_stride), md(1024, self.output_stride))
            self.args.train_crop_size = (cy, cx)
            logging.warning('--train_crop_size is not set. It will be set as (%d, %d).', cy, cx)
        else:
            cy, cx = (
                md(self.args.train_crop_size[0], self.output_stride), md(self.args.train_crop_size[1], self.output_stride))
        transform = ft.Compose([
            ft.ToTensor(),
            ft.Resize((cy, cx))
        ])

        dataset_name = 'sintel'
        if len(args) > 0 and args[0] in ['chairs2']:
            dataset_name = args[0]

        if dataset_name == 'sintel':
            dataset = SintelDataset(
                self.args.mpi_sintel_root_dir, split='trainval', pass_names='clean', transform=transform,
                get_occlusion_mask=False)
        elif dataset_name == 'chairs2':
            dataset = FlyingChairs2Dataset(
                self.args.flying_chairs2_root_dir, split='trainval', transform=transform, add_reverse=False,
                get_occlusion_mask=True, get_motion_boundary_mask=True, get_backward=True)

        dataset.img_paths = dataset.img_paths[:1]
        dataset.flow_paths = dataset.flow_paths[:1]
        dataset.occ_paths = dataset.occ_paths[:1]
        dataset.mb_paths = dataset.mb_paths[:1]
        dataset.flow_b_paths = dataset.flow_b_paths[:1]
        dataset.occ_b_paths = dataset.occ_b_paths[:1]
        dataset.mb_b_paths = dataset.mb_b_paths[:1]
        dataset.metadata = dataset.metadata[:1]

        return dataset



"""Operations to perform image augmentations for optical flow.

Some operations are adapted from the following sources:

- FlowNetPytorch: https://github.com/ClementPinard/FlowNetPytorch

- RAFT: https://github.com/princeton-vl/RAFT/

- flow-transforms-pytorch: https://github.com/hmorimitsu/flow-transforms-pytorch
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from collections.abc import KeysView
import random
from typing import Dict, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms as tt


class Compose(object):
    """Similar to torchvision Compose. Applies a series of transforms from the input list in sequence."""

    def __init__(
        self,
        transforms_list: Sequence[object]
    ) -> None:
        """Initialize Compose.

        Parameters
        ----------
        transforms_list : Sequence[object]
            A sequence of transforms to be applied.
        """
        self.transforms_list = transforms_list

    def __call__(
        self,
        inputs: Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
            Each element of the dict is either a single 3D HWC image or a list of images.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        for t in self.transforms_list:
            inputs = t(inputs)
        return inputs


class ToTensor(object):
    """Converts a 4D numpy.ndarray or a list of 3D numpy.ndarrays into a 4D torch.Tensor.

    If an input is of type uint8, then it is converted to float and its values are divided by 255.
    """

    def __init__(
        self,
        fp16: bool = False,
        device: Union[str, torch.device] = 'cpu',
        use_keys: Optional[Union[KeysView, Sequence[str]]] = None,
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None
    ) -> None:
        """Initialize ToTensor.

        Parameters
        ----------
        fp16 : bool, default False
            If True, the tensors use have-precision floating point.
        device : Union[str, torch.device], default 'cpu'
            Name of the torch device where the tensors will be put in.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        self.dtype = torch.float16 if fp16 else torch.float32
        self.device = device
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys

    def __call__(
        self,
        inputs: Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
            Each element of the dict is either a single 3D HWC image or a list of images.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for k in valid_keys:
            v = inputs[k]
            if isinstance(v, list) or isinstance(v, tuple):
                v = np.stack(v)
                if len(v.shape) == 3:
                    v = v[:, :, :, None]

            if len(v.shape) == 2:
                v = v[None, :, :, None]
            elif len(v.shape) == 3:
                v = v[None]

            if v.dtype == np.uint8:
                v = v.astype(np.float32) / 255.0
            v = v.transpose(0, 3, 1, 2)
            inputs[k] = torch.from_numpy(v).to(device=self.device, dtype=self.dtype)
        return inputs


class ColorJitter(tt.ColorJitter):
    """Randomly apply color transformations only to the images.

    If asymmetric_prob == 0, then the same transform is applied on all the images, otherwise, the transform for each image
    is randomly sampled independently.

    This is basically a wrapper for torchvision.transforms.ColorJitter.
    """

    def __init__(
        self,
        brightness: Union[float, Tuple[float, float]] = 0.0,
        contrast: Union[float, Tuple[float, float]] = 0.0,
        saturation: Union[float, Tuple[float, float]] = 0.0,
        hue: Union[float, Tuple[float, float]] = 0.0,
        asymmetric_prob: float = 0.0,
        use_keys: Optional[Union[KeysView, Sequence[str]]] = ('images',),
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None
    ) -> None:
        """Initialize ColorJitter.

        Parameters
        ----------
        brightness : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random brightness value.
        contrast : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random contrast value.
        saturation : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random saturation value.
        hue : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random hue value.
        asymmetric_prob : float, default 0.0
            Chance to apply an asymmetric transform, in which the parameters for transforming each image are sampled
            independently.
        use_keys : Optional[Union[KeysView, Sequence[str]]], default ['images']
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        super().__init__(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue)
        self.asymmetric_prob = asymmetric_prob
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys

    def __call__(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for k in valid_keys:
            v = inputs[k]
            if random.random() < self.asymmetric_prob:
                for i in range(len(v)):
                    inputs[k][i] = super().__call__(v[i])
            else:
                inputs[k] = super().__call__(v)
        return inputs


class GaussianNoise(object):
    """Applies random gaussian noise on the images."""

    def __init__(
        self,
        stdev: float = 0.0,
        use_keys: Optional[Union[KeysView, Sequence[str]]] = ('images',),
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None
    ) -> None:
        """Initialize GaussianNoise.

        Parameters
        ----------
        stdev : float, default 0.0
            The maximum standard deviation of the gaussian noise.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        self.stdev = stdev
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys

    def __call__(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for k in valid_keys:
            v = inputs[k]
            std = random.uniform(0.0, self.stdev)
            inputs[k] = (v + std * torch.randn(*v.shape, dtype=v.dtype, device=v.device)).clamp(0.0, 1.0)
        return inputs


class RandomPatchEraser(object):
    """Randomly covers a rectangular patch on the second image with noise, to simulate a pseudo-occlusion.

    The noise_type may be the mean or random. This transform erases patches ONLY FROM THE SECOND IMAGE.
    """

    def __init__(
        self,
        erase_prob: float = 0.0,
        num_patches: Union[int, Tuple[int, int]] = 1,
        patch_size: Union[Tuple[int, int], Tuple[int, int, int, int]] = (0, 0),
        noise_type: str = 'mean',
        use_keys: Optional[Union[KeysView, Sequence[str]]] = ('images',),
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None
    ) -> None:
        """Initialize RandomPatchEraser.

        Parameters
        ----------
        erase_prob : float, default 0.0
            Probability of applying the transformation.
        num_patches : Union[int, Tuple[int, int]], default 1
            Number of occlusion patches to generate. If it is a tuple, the number will be uniformly sampled from the interval.
        patch_size : Union[Tuple[int, int], Tuple[int, int, int, int]], default (0, 0)
            Range of the size of the occlusion patches. If it is a tuple with 2 elements, then both sides are sampled from the
            same interval. Otherwise, different intervals can be specified for each side as (hmin, hmax, wmin, wmax).
        noise_type : str, default 'mean'
            How to fill the occlusion patch. It can be either with the image 'mean' or with random 'noise'.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        self.erase_prob = erase_prob
        self.noise_type = noise_type
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys
        self.num_patches = num_patches
        if not (isinstance(num_patches, tuple) or isinstance(num_patches, list)):
            self.num_patches = (num_patches, num_patches)
        self.patch_size = patch_size
        if len(patch_size) == 2:
            self.patch_size = (patch_size[0], patch_size[1], patch_size[0], patch_size[1])

    def __call__(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        if random.random() < self.erase_prob:
            valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
            for k in valid_keys:
                im2 = inputs[k][1]
                c, h, w = im2.shape
                if self.noise_type == 'mean':
                    mean_color = im2.reshape(c, -1).mean(dim=1)
                for _ in range(random.randint(self.num_patches[0], self.num_patches[1])):
                    hp = random.randint(self.patch_size[0], self.patch_size[1])
                    wp = random.randint(self.patch_size[2], self.patch_size[3])
                    if hp > 0 and wp > 0:
                        yp = random.randint(0, h-hp)
                        xp = random.randint(0, w-wp)
                        if self.noise_type == 'mean':
                            noise = mean_color[:, None, None].repeat(1, hp, wp)
                        else:
                            im_min = im2.min()
                            im_max = im2.max()
                            im_inter = im_max - im_min
                            noise = im_inter*torch.rand(c, hp, wp, dtype=im2.dtype, device=im2.device) + im_min
                        im2[:, yp:yp+hp, xp:xp+wp] = noise
        return inputs


class RandomFlip(object):
    """Randomly horizontally and vertically flips the inputs.

    If asymmetric_prob > 0, then each input of the sequence may be flipped differently.
    """

    def __init__(
        self,
        hflip_prob: float = 0.0,
        vflip_prob: float = 0.0,
        asymmetric_prob: float = 0.0,
        use_keys: Optional[Union[KeysView, Sequence[str]]] = None,
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
        image_keys: Union[KeysView, Sequence[str]] = ('images',),
        flow_keys: Union[KeysView, Sequence[str]] = ('flows', 'flows_b')
    ) -> None:
        """Initialize RandomFlip.

        Parameters
        ----------
        hflip_prob : float, default 0.0
            Probability of applying a horizontal flip.
        vflip_prob : float, default 0.0
            Probability of applying a vertical flip.
        asymmetric_prob : float, default 0.0
            Chance to apply an asymmetric transform, in which the parameters for transforming each image are sampled
            independently.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        image_keys : Union[KeysView, Sequence[str]], ['images']
            Indicate which of the input keys correspond to image tensors.
        flow_keys : Union[KeysView, Sequence[str]], ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        """
        self.flip_probs = [hflip_prob, vflip_prob]
        self.asymmetric_prob = asymmetric_prob
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys
        self.image_keys = list(image_keys)
        self.flow_keys = list(flow_keys)

    def __call__(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for iorient in range(2):
            if self.asymmetric_prob < 1e-5:
                if random.random() < self.flip_probs[iorient]:
                    inputs = self._flip_inputs(inputs, iorient == 0, valid_keys)
            else:
                is_flips = [random.random() < self.flip_probs[iorient] for _ in range(inputs[self.image_keys[0]].shape[0])]
                for i in range(inputs[self.flow_keys[0]].shape[0]):
                    if is_flips[i]:
                        inputs = self._flip_inputs(inputs, iorient == 0, valid_keys, ibatch=i)
                    if is_flips[i] != is_flips[i+1]:
                        for fk in self.flow_keys:
                            inputs[fk][i] = self._mirror_flow(inputs[fk][i], iorient == 0)
                if is_flips[-1]:
                    for ik in self.image_keys:
                        inputs = self._flip_inputs(inputs, iorient == 0, inputs_keys=[ik], ibatch=-1)

        return inputs

    def _flip_inputs(
        self,
        inputs: Dict[str, torch.Tensor],
        is_hflip: bool,
        valid_keys: Optional[Sequence[str]] = None,
        ibatch: Optional[int] = None
    ) -> Dict[str, torch.Tensor]:
        """Flips all inputs horizontally or vertically.

        This function properly adjust the flow values after the flipping.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be flipped. Each element is a 4D tensor NCHW.
        is_hflip : bool
            If True, performs a horizontal flip, otherwise, performs a vertical flip.
        valid_keys : Optional[Sequence[str]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed.
        ibatch : Optional[int], optional
            If ibatch is specified, then only one element of the batch is flipped.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs flipped by this operation.
        """
        if is_hflip:
            iinp = 3
            iflow = 0
        else:
            iinp = 2
            iflow = 1

        if valid_keys is None:
            valid_keys = list(inputs.keys())
        for k in valid_keys:
            if ibatch is None:
                inputs[k] = torch.flip(inputs[k], [iinp])
                if k == 'flows':
                    inputs[k][:, iflow] *= -1
            else:
                inputs[k][ibatch] = torch.flip(inputs[k][ibatch], [iinp-1])
                if k == 'flows':
                    inputs[k][ibatch, iflow] *= -1
        return inputs

    def _mirror_flow(
        self,
        flow: torch.Tensor,
        is_hflip: bool
    ) -> torch.Tensor:
        """Reflects the flow along the center line of the image.

        This function is used when an asymmetric flip happens (one image flips, but the next does not, or vice-versa).

        Parameters
        ----------
        flow : torch.Tensor
            A 3D tensor CHW.
        is_hflip : bool
            If True, performs a horizontal flip, otherwise, performs a vertical flip.

        Returns
        -------
        torch.Tensor
            The mirrored flow.
        """
        grid = torch.meshgrid(torch.arange(flow.shape[1]), torch.arange(flow.shape[2]))
        grid = torch.stack(grid[::-1]).float()
        if is_hflip:
            mean_coord = (flow.shape[2]-1) / 2.0
            flow[0] = 2*(mean_coord-grid[0]) - flow[0]
        else:
            mean_coord = (flow.shape[1]-1) / 2.0
            flow[1] = 2*(mean_coord-grid[1]) - flow[1]
        return flow


class RandomScaleAndCrop(object):
    """Applies first random scale and then random crop to the inputs.

    The scale is adjusted so that it is not smaller than the crop size.
    If min_pool_binary is True, then inputs[binary_keys]
    are interpolated with min pooling, otherwise with bilinear interpolation.

    The scale calculation is composed of 2 main stages:

    1. A random major scale is sampled. The major scale defines the global scale applied to
       all images and dimensions. The major scale is calculated as:

       ms = 2 ** random.uniform(major_scale[0], major_scale[1])

    2. A random minor space scale is sampled. The space scale dictates the variation
       in scale applied to the width and height of each image. The space scale is
       calculated as:

       ssh = 2 ** random.uniform(space_scale[0], space_scale[1])

       ssw = 2 ** random.uniform(space_scale[2], space_scale[3]).

       If len(space_scale) == 2, then ssw also uses space_scale[0] and space_scale[1].

    The final scale applied to all inputs is:

        scale_height = ms * ssh

        scale_width = ms * ssw

    If time_scale is provided, then a third scale is sampled before computing the final scale.
    The time_scale is sampled independently for each element of a sequence. This allows,
    for example, for the first image have a different scale then the second one.
    The time scales tsh and tsw are calculated as the space scales ssh and ssw.
    With time scales, the final scales are calculated as:

        scale_height_time_t = ms * ssh * tsh_t

        scale_width_time_t = ms * ssw * tsw_t
    """

    def __init__(
        self,
        crop_size: Optional[Tuple[int, int]] = None,
        major_scale: Tuple[float, float] = (0.0, 0.0),
        space_scale: Union[Tuple[float, float], Tuple[float, float, float, float]] = (0.0, 0.0),
        time_scale: Union[Tuple[float, float], Tuple[float, float, float, float]] = (0.0, 0.0),
        min_pool_binary: bool = True,
        binary_keys: Union[KeysView, Sequence[str]] = ('mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b'),
        flow_keys: Union[KeysView, Sequence[str]] = ('flows', 'flows_b'),
        occlusion_keys: Union[KeysView, Sequence[str]] = ('occs', 'occs_b')
    ) -> None:
        """Initialize RandomScaleAndCrop.

        Parameters
        ----------
        crop_size : Optional[Tuple[int, int]], optional
            If provided, crop the inputs to this size (h, w).
        major_scale : Tuple[float, float], default (0.0, 0.0)
            The range of the major scale. See the class description for more details.
        space_scale : Union[Tuple[float, float], Tuple[float, float, float, float]], default (0.0, 0.0)
            The range of the minor scale. See the class description for more details.
        time_scale : Union[Tuple[float, float], Tuple[float, float, float, float]], default (0.0, 0.0)
            NOTE: Currently not implemented. The range of the time scale. See the class description for more details.
        min_pool_binary : bool, default True
            If True, min pooling is applied on binary inputs after the resizing. This ensures: 1. that they remain binary, and
            2. only pixels which were resized from patches containing only ones remain one.
        binary_keys : Union[KeysView, Sequence[str]], default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
            Indicate which of the input keys correspond to binary tensors.
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        """
        self.crop_size = crop_size
        self.min_pool_binary = min_pool_binary
        self.major_scale = major_scale
        if len(major_scale) == 2:
            self.major_scale = (major_scale[0], major_scale[1], major_scale[0], major_scale[1])
        self.space_scale = space_scale
        if len(space_scale) == 2:
            self.space_scale = (space_scale[0], space_scale[1], space_scale[0], space_scale[1])
        self.time_scale = time_scale
        if len(time_scale) == 2:
            self.time_scale = (time_scale[0], time_scale[1], time_scale[0], time_scale[1])

        self.use_time_scale = (abs(self.time_scale[1]-self.time_scale[0]) > 1e-5
                               and abs(self.time_scale[3]-self.time_scale[2]) > 1e-5)
        self.binary_keys = list(binary_keys)
        self.flow_keys = list(flow_keys)
        self.occlusion_keys = list(occlusion_keys)

    def __call__(  # noqa: C901
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.

        Raises
        ------
        NotImplementedError
            If trying to use time scale.
        """
        h, w = inputs[self.flow_keys[0]].shape[2:4]
        major_scale = 2 ** random.uniform(self.major_scale[0], self.major_scale[1])
        space_scales = (2 ** random.uniform(self.space_scale[0], self.space_scale[1]),
                        2 ** random.uniform(self.space_scale[2], self.space_scale[3]))
        if self.use_time_scale:
            raise NotImplementedError()
        else:
            min_size = self.crop_size
            if min_size is None:
                min_size = (1, 1)
            scaled_size = (max(min_size[0], int(h*major_scale*space_scales[0])),
                           max(min_size[1], int(w*major_scale*space_scales[1])))
            if self.crop_size is not None:
                y_crop = random.randint(0, scaled_size[0]-self.crop_size[0])
                x_crop = random.randint(0, scaled_size[1]-self.crop_size[1])
            for k, v in inputs.items():
                v = F.interpolate(v, size=scaled_size, mode='bilinear', align_corners=True)
                if self.min_pool_binary and k in self.binary_keys:
                    # Pseudo min pooling
                    v[v < 0.999] = 0
                if self.crop_size is not None:
                    v = v[:, :, y_crop:y_crop+self.crop_size[0], x_crop:x_crop+self.crop_size[1]]

                if k in self.flow_keys:
                    scale_mult = [float(scaled_size[1]) / w, float(scaled_size[0]) / h]
                    scale_mult = torch.from_numpy(np.array(scale_mult)).to(dtype=v.dtype, device=v.device)[None, :, None, None]
                    v = v * scale_mult
                inputs[k] = v

            # Update occlusion masks for out-of-bounds flows
            for k, v in inputs.items():
                try:
                    i = self.occlusion_keys.index(k)
                    inputs[k] = _update_oob_flows(v, inputs[self.flow_keys[i]])
                except ValueError:
                    pass
        return inputs


class RandomTranslate(object):
    """Creates a translation between images by applying a random alternated crop on the sequence of inputs.

    A translation value t is randomly selected first. Then, the first image is cropped by a box translated by t.
    The second image will be cropped by a reversed translation -t. The third will be cropped by t again, and so on...
    """

    def __init__(
        self,
        translation: Union[int, Tuple[int, int]] = 0,
        flow_keys: Union[KeysView, Sequence[str]] = ('flows', 'flows_b'),
        occlusion_keys: Union[KeysView, Sequence[str]] = ('occs', 'occs_b')
    ) -> None:
        """Initialize RandomTranslate.

        Parameters
        ----------
        translation : Union[int, Tuple[int, int]], default 0
            Maximum translation (in pixels) to be applied to the inputs. If a tuple, it corresponds to the maximum in the
            (y, x) axes.
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        """
        self.translation = translation
        if not isinstance(translation, tuple) or isinstance(translation, list):
            self.translation = (translation, translation)
        self.flow_keys = flow_keys
        self.occlusion_keys = occlusion_keys

    def __call__(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        _, _, h, w = inputs[self.flow_keys[0]].shape
        th, tw = self.translation
        tw = random.randint(-tw, tw)
        th = random.randint(-th, th)
        if tw == 0 and th == 0:
            return inputs

        trans_inputs = {k: torch.empty_like(v[:, :, :h-abs(th), :w-abs(tw)]) for k, v in inputs.items()}

        # Translate: 0: even indexed inputs, 1: odd indexed inputs
        for t in range(2):
            if t == 0:
                ftw = tw
                fth = th
            else:
                ftw = -tw
                fth = -th
            x1, x2 = max(0, ftw), min(w+ftw, w)
            y1, y2 = max(0, fth), min(h+fth, h)
            for k, v in inputs.items():
                trans_inputs[k][t::2] = v[t::2, :, y1:y2, x1:x2]
                if k in self.flow_keys:
                    trans_inputs[k][t::2, 0] += ftw
                    trans_inputs[k][t::2, 1] += fth

        # Update occlusion masks for out-of-bounds flows
        for k, v in trans_inputs.items():
            try:
                i = self.occlusion_keys.index(k)
                trans_inputs[k] = _update_oob_flows(v, trans_inputs[self.flow_keys[i]])
            except ValueError:
                pass

        return trans_inputs


class RandomRotate(object):
    """Applies random rotation to the inputs.

    The inputs are rotated around the center of the image. First all inputs are rotated by the same random major `angle`.
    Then, another random angle a is sampled according to `diff_angle`. The first image will be rotated by a. The second image
    will be rotated by a reversed angle -a. The third will be rotated by a again, and so on...
    """

    def __init__(
        self,
        angle: float = 0.0,
        diff_angle: float = 0.0,
        min_pool_binary: bool = True,
        flow_keys: Union[KeysView, Sequence[str]] = ('flows', 'flows_b'),
        occlusion_keys: Union[KeysView, Sequence[str]] = ('occs', 'occs_b'),
        valid_keys: Union[KeysView, Sequence[str]] = ('valids', 'valids_b'),
        binary_keys: Union[KeysView, Sequence[str]] = ('mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b')
    ) -> None:
        """Initialize RandomRotate.

        Parameters
        ----------
        angle : float, default 0.0
            The maximum absolute value to sample the major angle from.
        diff_angle : float, default 0.0
            The maximum absolute value to sample the angle difference between consecutive images.
        min_pool_binary : bool, default True
            If True, min pooling is applied on binary inputs after the resizing. This ensures: 1. that they remain binary, and
            2. only pixels which were resized from patches containing only ones remain one.
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        valid_keys : Union[KeysView, Sequence[str]], default ['valids', 'valids_b']
            Indicate which of the input keys correspond to valid mask tensors.
        binary_keys : Union[KeysView, Sequence[str]], default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
            Indicate which of the input keys correspond to binary tensors.
        """
        self.angle = angle
        self.diff_angle = diff_angle
        self.min_pool_binary = min_pool_binary
        self.flow_keys = flow_keys
        self.occlusion_keys = occlusion_keys
        self.valid_keys = valid_keys
        self.binary_keys = binary_keys

    def __call__(  # noqa: C901
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        major_angle = random.uniform(-self.angle, self.angle)
        inter_angle = random.uniform(-self.diff_angle, self.diff_angle)

        input_tensor = inputs[self.flow_keys[0]]
        b, _, h, w = input_tensor.shape

        def generate_rotation_grid(
            rot_angle: float,
            batch_size: int,
            dtype: torch.dtype,
            device: torch.device
        ) -> torch.Tensor:
            vy, vx = torch.meshgrid(torch.arange(h), torch.arange(w))
            vx = vx.type(dtype)
            vy = vy.type(dtype)
            vx = vx.to(device)
            vy = vy.to(device)
            vx -= (w-1.0)/2.0
            vy -= (h-1.0)/2.0
            angle_rad = rot_angle*2*np.pi / 360
            rotx = np.cos(angle_rad)*vx - np.sin(angle_rad)*vy
            roty = np.sin(angle_rad)*vx + np.cos(angle_rad)*vy
            rotx = rotx / ((w-1)/2)
            roty = roty / ((h-1)/2)
            rot_grid = torch.stack((rotx, roty), dim=2)[None]
            rot_grid = rot_grid.repeat(batch_size, 1, 1, 1)
            return rot_grid

        def generate_rotation_matrix(
            rot_angle: float,
            batch_size: int,
            dtype: torch.dtype,
            device: torch.device
        ) -> torch.Tensor:
            vx, vy = torch.meshgrid(torch.arange(h), torch.arange(w))
            vx = vx.type(dtype)
            vy = vy.type(dtype)
            vx = vx.to(device)
            vy = vy.to(device)
            rotx = (vx - h/2.0) * (rot_angle*np.pi/180.0)
            roty = -(vy - w/2.0) * (rot_angle*np.pi/180.0)
            rot_mat = torch.stack((rotx, roty), dim=0)[None]
            rot_mat = rot_mat.repeat(batch_size, 1, 1, 1)
            return rot_mat

        def rotate_flow(
            flow: torch.Tensor,
            rot_angle: float
        ) -> torch.Tensor:
            angle_rad = rot_angle*2*np.pi / 360
            rot_flow = flow.clone()
            rot_flow[:, 0] = (
                np.cos(angle_rad)*flow[:, 0] + np.sin(angle_rad)*flow[:, 1])
            rot_flow[:, 1] = (
                -np.sin(angle_rad)*flow[:, 0] + np.cos(angle_rad)*flow[:, 1])
            return rot_flow

        rot_mat = generate_rotation_matrix(inter_angle, b//2+1, input_tensor.dtype, input_tensor.device)
        for t in range(2):
            if t == 0:
                inangle = -inter_angle
                rmat = rot_mat
            else:
                inangle = inter_angle
                rmat = -rot_mat
            angle = major_angle + inangle / 2
            num_flows = input_tensor[t::2].shape[0]
            num_images = num_flows + 1
            rot_grid = generate_rotation_grid(angle, num_images, input_tensor.dtype, input_tensor.device)
            for k, v in inputs.items():
                if k in self.flow_keys:
                    v[t::2] += rmat[:num_flows]

                v[t::2] = F.grid_sample(v[t::2], rot_grid[:v[t::2].shape[0]], mode='bilinear', align_corners=True)

                if k in self.flow_keys:
                    v[t::2] = rotate_flow(v[t::2], angle)
                elif self.min_pool_binary and k in self.binary_keys:
                    # Pseudo min pooling
                    v[v < 0.999] = 0
                inputs[k] = v

        # Update occlusion masks for out-of-bounds flows
        for k, v in inputs.items():
            try:
                i = self.occlusion_keys.index(k)
                v = _update_oob_flows(v, inputs[self.flow_keys[i]])
                inputs[k] = v
            except ValueError:
                pass

        return inputs


class Resize(object):
    """Resize the image to a given size or scale.

    Size is checked first, if any of its values is zero, then scale is used.
    """

    def __init__(
        self,
        size: Tuple[int, int] = (0, 0),
        scale: float = 1.0,
        min_pool_binary: bool = True,
        binary_keys: Union[KeysView, Sequence[str]] = ('mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b'),
        flow_keys: Union[KeysView, Sequence[str]] = ('flows', 'flows_b')
    ) -> None:
        """Initialize Resize.

        Parameters
        ----------
        size : Tuple[int, int], default (0, 0)
            The target size to resize the inputs. If it is zeros, then the scale will be used instead.
        scale : float, default 1.0
            The scale factor to resize the images. Only used if size is zeros.
        min_pool_binary : bool, default True
            If True, min pooling is applied on binary inputs after the resizing. This ensures: 1. that they remain binary, and
            2. only pixels which were resized from patches containing only ones remain one.
        binary_keys : Union[KeysView, Sequence[str]], default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
            Indicate which of the input keys correspond to binary tensors.
            [description], by default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        """
        self.size = size
        self.scale = scale
        self.min_pool_binary = min_pool_binary
        self.binary_keys = list(binary_keys)
        self.flow_keys = list(flow_keys)

    def __call__(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        h, w = inputs[list(inputs.keys())[0]].shape[2:4]
        if self.size is None or self.size[0] < 1 or self.size[1] < 1:
            self.size = (int(self.scale*h), int(self.scale*w))
        if self.size[0] != h or self.size[1] != w:
            for k, v in inputs.items():
                v = F.interpolate(v, size=self.size, mode='bilinear', align_corners=True)
                if self.min_pool_binary and k in self.binary_keys:
                    # Pseudo min pooling
                    v[v < 0.999] = 0

                if k in self.flow_keys:
                    scale_mult = [float(self.size[1]) / w, float(self.size[0]) / h]
                    scale_mult = torch.from_numpy(np.array(scale_mult)).to(dtype=v.dtype, device=v.device)[None, :, None, None]
                    v = v * scale_mult
                inputs[k] = v
        return inputs


def _get_valid_keys(
    inputs_keys: Union[KeysView, Sequence[str]],
    use_keys: Optional[Union[KeysView, Sequence[str]]],
    ignore_keys: Optional[Union[KeysView, Sequence[str]]]
) -> Union[KeysView, Sequence[str]]:
    """Get only the valid keys from the input.

    Basically, it return use_keys, if not None. Otherwise, return inputs_keys after removing the keys which are in ignore_keys.

    Parameters
    ----------
    inputs_keys : Union[KeysView, Sequence[str]]
        All the keys available as input.
    use_keys : Optional[Union[KeysView, Sequence[str]]]
        If not None, then just use these keys.
    ignore_keys : Optional[Union[KeysView, Sequence[str]]]
        If not None, remove these keys from the inputs_keys.

    Returns
    -------
    Union[KeysView, Sequence[str]]
        The keys remaining after the validity checks.
    """
    if use_keys is not None:
        return use_keys
    if ignore_keys is None:
        return inputs_keys
    return [k for k in inputs_keys if k not in ignore_keys]


def _update_oob_flows(
    occs: torch.Tensor,
    flows: torch.Tensor
) -> torch.Tensor:
    """Update occlusion maps to include flow which went out-of-bounds.

    Parameters
    ----------
    occs : torch.Tensor
        A 4D tensor NCHW of occlusion masks.
    flows : torch.Tensor
        A 4D tensor NCHW of optical flows.

    Returns
    -------
    torch.Tensor
        The updated occlusion masks. Flows which went out-of-bounds are marked as occluded.
    """
    grid = torch.meshgrid(
        torch.arange(flows.shape[2], dtype=flows.dtype, device=flows.device),
        torch.arange(flows.shape[3], dtype=flows.dtype, device=flows.device))
    grid = torch.stack(grid[::-1]).float()[None].repeat(flows.shape[0], 1, 1, 1)
    coords = flows + grid
    oob_occs = coords < 0
    oob_occs[:, 0] |= coords[:, 0] > flows.shape[3]
    oob_occs[:, 1] |= coords[:, 1] > flows.shape[2]
    oob_occs = oob_occs.max(dim=1, keepdim=True)[0].to(dtype=occs.dtype, device=occs.device)
    occs = torch.max(torch.stack([occs, oob_occs], dim=0), dim=0)[0]
    return occs



'''
Portions of this code copyright 2017, Clement Pinard
'''

# freda (todo) : adversarial loss 

import torch
import torch.nn as nn
import math

def EPE(input_flow, target_flow):
    return torch.norm(target_flow-input_flow,p=2,dim=1).mean()

class L1(nn.Module):
    def __init__(self):
        super(L1, self).__init__()
    def forward(self, output, target):
        lossvalue = torch.abs(output - target).mean()
        return lossvalue

class L2(nn.Module):
    def __init__(self):
        super(L2, self).__init__()
    def forward(self, output, target):
        lossvalue = torch.norm(output-target,p=2,dim=1).mean()
        return lossvalue

class L1Loss(nn.Module):
    def __init__(self):
        super(L1Loss, self).__init__()
        self.loss = L1()
        self.loss_labels = ['L1', 'EPE']

    def forward(self, output, target):
        output = output['flow_preds']
        target = target['flows'][:, 0]
        lossvalue = self.loss(output, target)
        return lossvalue

class L2Loss(nn.Module):
    def __init__(self):
        super(L2Loss, self).__init__()
        self.loss = L2()
        self.loss_labels = ['L2', 'EPE']

    def forward(self, output, target):
        output = output['flow_preds']
        target = target['flows'][:, 0]
        lossvalue = self.loss(output, target)
        return lossvalue

class MultiScale(nn.Module):
    def __init__(self, startScale = 4, numScales = 5, l_weight= 0.32, norm= 'L1'):
        super(MultiScale,self).__init__()

        self.startScale = startScale
        self.numScales = numScales
        self.loss_weights = torch.FloatTensor([(l_weight / 2 ** scale) for scale in range(self.numScales)])
        self.l_type = norm
        self.div_flow = 0.05
        assert(len(self.loss_weights) == self.numScales)

        if self.l_type == 'L1':
            self.loss = L1()
        else:
            self.loss = L2()

        self.multiScales = [nn.AvgPool2d(self.startScale * (2**scale), self.startScale * (2**scale)) for scale in range(self.numScales)]
        self.loss_labels = ['MultiScale-'+self.l_type, 'EPE']

    def forward(self, output, target):
        output = output['flow_preds']
        target = target['flows'][:, 0]
        lossvalue = 0

        if type(output) is tuple or type(output) is list:
            target = self.div_flow * target
            for i, flow_pred_ in enumerate(output):
                target_ = self.multiScales[i](target)
                lossvalue += self.loss_weights[i]*self.loss(flow_pred_, target_)
        else:
            lossvalue += self.loss(output, target)
        return lossvalue



'''
Portions of this code copyright 2017, Clement Pinard
'''

from argparse import ArgumentParser, Namespace
from pathlib import Path

import numpy as np
try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import IterSpatialCorrelationSampler as SpatialCorrelationSampler
import torch
import torch.nn as nn

from ..base_model.base_model import BaseModel
from ..flownet.losses import MultiScale


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):   
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, 
                        padding=padding, dilation=dilation, bias=True),
            nn.LeakyReLU(0.1))


def predict_flow(in_planes):
    return nn.Conv2d(in_planes,2,kernel_size=3,stride=1,padding=1,bias=True)


def deconv(in_planes, out_planes, kernel_size=4, stride=2, padding=1):
    return nn.ConvTranspose2d(in_planes, out_planes, kernel_size, stride, padding, bias=True)


class PWCNet(BaseModel):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcnet-things-6a2e540b.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcnet-sintel-533815e5.ckpt'
    }

    def __init__(self,
                 args: Namespace):
        super(PWCNet, self).__init__(
            args=args,
            loss_fn=MultiScale(
                startScale=args.loss_start_scale,
                numScales=args.loss_num_scales,
                l_weight=args.loss_base_weight,
                norm=args.loss_norm
            ),
            output_stride=64)

        self.conv1a  = conv(3,   16, kernel_size=3, stride=2)
        self.conv1aa = conv(16,  16, kernel_size=3, stride=1)
        self.conv1b  = conv(16,  16, kernel_size=3, stride=1)
        self.conv2a  = conv(16,  32, kernel_size=3, stride=2)
        self.conv2aa = conv(32,  32, kernel_size=3, stride=1)
        self.conv2b  = conv(32,  32, kernel_size=3, stride=1)
        self.conv3a  = conv(32,  64, kernel_size=3, stride=2)
        self.conv3aa = conv(64,  64, kernel_size=3, stride=1)
        self.conv3b  = conv(64,  64, kernel_size=3, stride=1)
        self.conv4a  = conv(64,  96, kernel_size=3, stride=2)
        self.conv4aa = conv(96,  96, kernel_size=3, stride=1)
        self.conv4b  = conv(96,  96, kernel_size=3, stride=1)
        self.conv5a  = conv(96, 128, kernel_size=3, stride=2)
        self.conv5aa = conv(128,128, kernel_size=3, stride=1)
        self.conv5b  = conv(128,128, kernel_size=3, stride=1)
        self.conv6aa = conv(128,196, kernel_size=3, stride=2)
        self.conv6a  = conv(196,196, kernel_size=3, stride=1)
        self.conv6b  = conv(196,196, kernel_size=3, stride=1)

        self.leakyRELU = nn.LeakyReLU(0.1)

        self.corr = SpatialCorrelationSampler(kernel_size=1, patch_size=2*self.args.md+1, padding=0)
        
        nd = (2*self.args.md+1)**2
        dd = np.cumsum([128,128,96,64,32])

        od = nd
        self.conv6_0 = conv(od,      128, kernel_size=3, stride=1)
        self.conv6_1 = conv(od+dd[0],128, kernel_size=3, stride=1)
        self.conv6_2 = conv(od+dd[1],96,  kernel_size=3, stride=1)
        self.conv6_3 = conv(od+dd[2],64,  kernel_size=3, stride=1)
        self.conv6_4 = conv(od+dd[3],32,  kernel_size=3, stride=1)        
        self.predict_flow6 = predict_flow(od+dd[4])
        self.deconv6 = deconv(2, 2, kernel_size=4, stride=2, padding=1) 
        self.upfeat6 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1) 
        
        od = nd+128+4
        self.conv5_0 = conv(od,      128, kernel_size=3, stride=1)
        self.conv5_1 = conv(od+dd[0],128, kernel_size=3, stride=1)
        self.conv5_2 = conv(od+dd[1],96,  kernel_size=3, stride=1)
        self.conv5_3 = conv(od+dd[2],64,  kernel_size=3, stride=1)
        self.conv5_4 = conv(od+dd[3],32,  kernel_size=3, stride=1)
        self.predict_flow5 = predict_flow(od+dd[4]) 
        self.deconv5 = deconv(2, 2, kernel_size=4, stride=2, padding=1) 
        self.upfeat5 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1) 
        
        od = nd+96+4
        self.conv4_0 = conv(od,      128, kernel_size=3, stride=1)
        self.conv4_1 = conv(od+dd[0],128, kernel_size=3, stride=1)
        self.conv4_2 = conv(od+dd[1],96,  kernel_size=3, stride=1)
        self.conv4_3 = conv(od+dd[2],64,  kernel_size=3, stride=1)
        self.conv4_4 = conv(od+dd[3],32,  kernel_size=3, stride=1)
        self.predict_flow4 = predict_flow(od+dd[4]) 
        self.deconv4 = deconv(2, 2, kernel_size=4, stride=2, padding=1) 
        self.upfeat4 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1) 
        
        od = nd+64+4
        self.conv3_0 = conv(od,      128, kernel_size=3, stride=1)
        self.conv3_1 = conv(od+dd[0],128, kernel_size=3, stride=1)
        self.conv3_2 = conv(od+dd[1],96,  kernel_size=3, stride=1)
        self.conv3_3 = conv(od+dd[2],64,  kernel_size=3, stride=1)
        self.conv3_4 = conv(od+dd[3],32,  kernel_size=3, stride=1)
        self.predict_flow3 = predict_flow(od+dd[4]) 
        self.deconv3 = deconv(2, 2, kernel_size=4, stride=2, padding=1) 
        self.upfeat3 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1) 
        
        od = nd+32+4
        self.conv2_0 = conv(od,      128, kernel_size=3, stride=1)
        self.conv2_1 = conv(od+dd[0],128, kernel_size=3, stride=1)
        self.conv2_2 = conv(od+dd[1],96,  kernel_size=3, stride=1)
        self.conv2_3 = conv(od+dd[2],64,  kernel_size=3, stride=1)
        self.conv2_4 = conv(od+dd[3],32,  kernel_size=3, stride=1)
        self.predict_flow2 = predict_flow(od+dd[4]) 

        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight.data, mode='fan_in')
                if m.bias is not None:
                    m.bias.data.zero_()
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear', align_corners=True)

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=20.0)
        parser.add_argument('--md', type=int, default=4)
        parser.add_argument('--loss_start_scale', type=float, default=4)
        parser.add_argument('--loss_num_scales', type=int, default=5)
        parser.add_argument('--loss_base_weight', type=float, default=0.32)
        parser.add_argument('--loss_norm', type=str, default='L2')
        return parser

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid 
        xx = torch.arange(0, W).view(1,-1).repeat(H,1)
        yy = torch.arange(0, H).view(-1,1).repeat(1,W)
        xx = xx.view(1,1,H,W).repeat(B,1,1,1)
        yy = yy.view(1,1,H,W).repeat(B,1,1,1)
        grid = torch.cat((xx,yy),1).float()

        if x.is_cuda:
            grid = grid.to(dtype=x.dtype, device=x.device)
        vgrid = grid + flo

        # scale grid to [-1,1] 
        vgrid[:,0,:,:] = 2.0*vgrid[:,0,:,:].clone() / max(W-1,1)-1.0
        vgrid[:,1,:,:] = 2.0*vgrid[:,1,:,:].clone() / max(H-1,1)-1.0

        vgrid = vgrid.permute(0,2,3,1)        
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.ones(x.size()).to(dtype=x.dtype, device=x.device)
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)
        
        mask[mask<0.9999] = 0
        mask[mask>0] = 1
        
        return output*mask


    def forward(self, inputs):
        im1 = inputs['images'][:, 0]
        im2 = inputs['images'][:, 1]
        
        c11 = self.conv1b(self.conv1aa(self.conv1a(im1)))
        c21 = self.conv1b(self.conv1aa(self.conv1a(im2)))
        c12 = self.conv2b(self.conv2aa(self.conv2a(c11)))
        c22 = self.conv2b(self.conv2aa(self.conv2a(c21)))
        c13 = self.conv3b(self.conv3aa(self.conv3a(c12)))
        c23 = self.conv3b(self.conv3aa(self.conv3a(c22)))
        c14 = self.conv4b(self.conv4aa(self.conv4a(c13)))
        c24 = self.conv4b(self.conv4aa(self.conv4a(c23)))
        c15 = self.conv5b(self.conv5aa(self.conv5a(c14)))
        c25 = self.conv5b(self.conv5aa(self.conv5a(c24)))
        c16 = self.conv6b(self.conv6a(self.conv6aa(c15)))
        c26 = self.conv6b(self.conv6a(self.conv6aa(c25)))


        corr6 = self.corr(c16, c26)
        corr6 = corr6.view(corr6.shape[0], -1, corr6.shape[3], corr6.shape[4])
        corr6 = corr6 / c16.shape[1]
        corr6 = self.leakyRELU(corr6)


        x = torch.cat((self.conv6_0(corr6), corr6),1)
        x = torch.cat((self.conv6_1(x), x),1)
        x = torch.cat((self.conv6_2(x), x),1)
        x = torch.cat((self.conv6_3(x), x),1)
        x = torch.cat((self.conv6_4(x), x),1)
        flow6 = self.predict_flow6(x)
        up_flow6 = self.deconv6(flow6)
        up_feat6 = self.upfeat6(x)

        
        warp5 = self.warp(c25, up_flow6*0.625)
        corr5 = self.corr(c15, warp5)
        corr5 = corr5.view(corr5.shape[0], -1, corr5.shape[3], corr5.shape[4])
        corr5 = corr5 / c15.shape[1]
        corr5 = self.leakyRELU(corr5)
        x = torch.cat((corr5, c15, up_flow6, up_feat6), 1)
        x = torch.cat((self.conv5_0(x), x),1)
        x = torch.cat((self.conv5_1(x), x),1)
        x = torch.cat((self.conv5_2(x), x),1)
        x = torch.cat((self.conv5_3(x), x),1)
        x = torch.cat((self.conv5_4(x), x),1)
        flow5 = self.predict_flow5(x)
        up_flow5 = self.deconv5(flow5)
        up_feat5 = self.upfeat5(x)

       
        warp4 = self.warp(c24, up_flow5*1.25)
        corr4 = self.corr(c14, warp4)
        corr4 = corr4.view(corr4.shape[0], -1, corr4.shape[3], corr4.shape[4])
        corr4 = corr4 / c14.shape[1]
        corr4 = self.leakyRELU(corr4)
        x = torch.cat((corr4, c14, up_flow5, up_feat5), 1)
        x = torch.cat((self.conv4_0(x), x),1)
        x = torch.cat((self.conv4_1(x), x),1)
        x = torch.cat((self.conv4_2(x), x),1)
        x = torch.cat((self.conv4_3(x), x),1)
        x = torch.cat((self.conv4_4(x), x),1)
        flow4 = self.predict_flow4(x)
        up_flow4 = self.deconv4(flow4)
        up_feat4 = self.upfeat4(x)


        warp3 = self.warp(c23, up_flow4*2.5)
        corr3 = self.corr(c13, warp3)
        corr3 = corr3.view(corr3.shape[0], -1, corr3.shape[3], corr3.shape[4])
        corr3 = corr3 / c13.shape[1]
        corr3 = self.leakyRELU(corr3)
        

        x = torch.cat((corr3, c13, up_flow4, up_feat4), 1)
        x = torch.cat((self.conv3_0(x), x),1)
        x = torch.cat((self.conv3_1(x), x),1)
        x = torch.cat((self.conv3_2(x), x),1)
        x = torch.cat((self.conv3_3(x), x),1)
        x = torch.cat((self.conv3_4(x), x),1)
        flow3 = self.predict_flow3(x)
        up_flow3 = self.deconv3(flow3)
        up_feat3 = self.upfeat3(x)


        warp2 = self.warp(c22, up_flow3*5.0)
        corr2 = self.corr(c12, warp2)
        corr2 = corr2.view(corr2.shape[0], -1, corr2.shape[3], corr2.shape[4])
        corr2 = corr2 / c12.shape[1]
        corr2 = self.leakyRELU(corr2)
        x = torch.cat((corr2, c12, up_flow3, up_feat3), 1)
        x = torch.cat((self.conv2_0(x), x),1)
        x = torch.cat((self.conv2_1(x), x),1)
        x = torch.cat((self.conv2_2(x), x),1)
        x = torch.cat((self.conv2_3(x), x),1)
        x = torch.cat((self.conv2_4(x), x),1)
        flow2 = self.predict_flow2(x)

        flow_up = self.upsample1(flow2*self.args.div_flow)

        outputs = {}
        if self.training:
            outputs['flow_preds'] = [flow2, flow3, flow4, flow5, flow6]
            outputs['flows'] = flow_up[:, None]
        else:
            outputs['flows'] = flow_up[:, None]
        return outputs


class PWCDCNet(PWCNet):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcdcnet-things-cc223701.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcdcnet-sintel-c7d08a46.ckpt'
    }

    def __init__(self,
                 args: Namespace):
        super(PWCDCNet, self).__init__(args=args)
        
        nd = (2*self.args.md+1)**2
        dd = np.cumsum([128,128,96,64,32])
        od = nd+32+4
        
        self.dc_conv1 = conv(od+dd[4], 128, kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc_conv2 = conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2)
        self.dc_conv3 = conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4)
        self.dc_conv4 = conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8)
        self.dc_conv5 = conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc_conv6 = conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc_conv7 = predict_flow(32)

        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight.data, mode='fan_in')
                if m.bias is not None:
                    m.bias.data.zero_()


    def forward(self, inputs):
        im1 = inputs['images'][:, 0]
        im2 = inputs['images'][:, 1]
        
        c11 = self.conv1b(self.conv1aa(self.conv1a(im1)))
        c21 = self.conv1b(self.conv1aa(self.conv1a(im2)))
        c12 = self.conv2b(self.conv2aa(self.conv2a(c11)))
        c22 = self.conv2b(self.conv2aa(self.conv2a(c21)))
        c13 = self.conv3b(self.conv3aa(self.conv3a(c12)))
        c23 = self.conv3b(self.conv3aa(self.conv3a(c22)))
        c14 = self.conv4b(self.conv4aa(self.conv4a(c13)))
        c24 = self.conv4b(self.conv4aa(self.conv4a(c23)))
        c15 = self.conv5b(self.conv5aa(self.conv5a(c14)))
        c25 = self.conv5b(self.conv5aa(self.conv5a(c24)))
        c16 = self.conv6b(self.conv6a(self.conv6aa(c15)))
        c26 = self.conv6b(self.conv6a(self.conv6aa(c25)))


        corr6 = self.corr(c16, c26)
        corr6 = corr6.view(corr6.shape[0], -1, corr6.shape[3], corr6.shape[4])
        corr6 = corr6 / c16.shape[1]
        corr6 = self.leakyRELU(corr6)


        x = torch.cat((self.conv6_0(corr6), corr6),1)
        x = torch.cat((self.conv6_1(x), x),1)
        x = torch.cat((self.conv6_2(x), x),1)
        x = torch.cat((self.conv6_3(x), x),1)
        x = torch.cat((self.conv6_4(x), x),1)
        flow6 = self.predict_flow6(x)
        up_flow6 = self.deconv6(flow6)
        up_feat6 = self.upfeat6(x)

        
        warp5 = self.warp(c25, up_flow6*0.625)
        corr5 = self.corr(c15, warp5)
        corr5 = corr5.view(corr5.shape[0], -1, corr5.shape[3], corr5.shape[4])
        corr5 = corr5 / c15.shape[1]
        corr5 = self.leakyRELU(corr5)
        x = torch.cat((corr5, c15, up_flow6, up_feat6), 1)
        x = torch.cat((self.conv5_0(x), x),1)
        x = torch.cat((self.conv5_1(x), x),1)
        x = torch.cat((self.conv5_2(x), x),1)
        x = torch.cat((self.conv5_3(x), x),1)
        x = torch.cat((self.conv5_4(x), x),1)
        flow5 = self.predict_flow5(x)
        up_flow5 = self.deconv5(flow5)
        up_feat5 = self.upfeat5(x)

       
        warp4 = self.warp(c24, up_flow5*1.25)
        corr4 = self.corr(c14, warp4)
        corr4 = corr4.view(corr4.shape[0], -1, corr4.shape[3], corr4.shape[4])
        corr4 = corr4 / c14.shape[1]
        corr4 = self.leakyRELU(corr4)
        x = torch.cat((corr4, c14, up_flow5, up_feat5), 1)
        x = torch.cat((self.conv4_0(x), x),1)
        x = torch.cat((self.conv4_1(x), x),1)
        x = torch.cat((self.conv4_2(x), x),1)
        x = torch.cat((self.conv4_3(x), x),1)
        x = torch.cat((self.conv4_4(x), x),1)
        flow4 = self.predict_flow4(x)
        up_flow4 = self.deconv4(flow4)
        up_feat4 = self.upfeat4(x)


        warp3 = self.warp(c23, up_flow4*2.5)
        corr3 = self.corr(c13, warp3)
        corr3 = corr3.view(corr3.shape[0], -1, corr3.shape[3], corr3.shape[4])
        corr3 = corr3 / c13.shape[1]
        corr3 = self.leakyRELU(corr3)
        

        x = torch.cat((corr3, c13, up_flow4, up_feat4), 1)
        x = torch.cat((self.conv3_0(x), x),1)
        x = torch.cat((self.conv3_1(x), x),1)
        x = torch.cat((self.conv3_2(x), x),1)
        x = torch.cat((self.conv3_3(x), x),1)
        x = torch.cat((self.conv3_4(x), x),1)
        flow3 = self.predict_flow3(x)
        up_flow3 = self.deconv3(flow3)
        up_feat3 = self.upfeat3(x)


        warp2 = self.warp(c22, up_flow3*5.0)
        corr2 = self.corr(c12, warp2)
        corr2 = corr2.view(corr2.shape[0], -1, corr2.shape[3], corr2.shape[4])
        corr2 = corr2 / c12.shape[1]
        corr2 = self.leakyRELU(corr2)
        x = torch.cat((corr2, c12, up_flow3, up_feat3), 1)
        x = torch.cat((self.conv2_0(x), x),1)
        x = torch.cat((self.conv2_1(x), x),1)
        x = torch.cat((self.conv2_2(x), x),1)
        x = torch.cat((self.conv2_3(x), x),1)
        x = torch.cat((self.conv2_4(x), x),1)
        flow2 = self.predict_flow2(x)
 
        x = self.dc_conv4(self.dc_conv3(self.dc_conv2(self.dc_conv1(x))))
        flow2 = flow2 + self.dc_conv7(self.dc_conv6(self.dc_conv5(x)))

        flow_up = self.upsample1(flow2*self.args.div_flow)

        outputs = {}
        if self.training:
            outputs['flow_preds'] = [flow2, flow3, flow4, flow5, flow6]
            outputs['flows'] = flow_up[:, None]
        else:
            outputs['flows'] = flow_up[:, None]
        return outputs



# Code taken from IRR: https://github.com/visinf/irr
# Licensed under the Apache 2.0 license (see LICENSE_IRR).

import torch
import torch.nn as nn
import torch.nn.functional as F


def get_grid(x):
    grid_H = torch.linspace(-1.0, 1.0, x.size(3)).view(1, 1, 1, x.size(3)).expand(x.size(0), 1, x.size(2), x.size(3))
    grid_V = torch.linspace(-1.0, 1.0, x.size(2)).view(1, 1, x.size(2), 1).expand(x.size(0), 1, x.size(2), x.size(3))
    grid = torch.cat([grid_H, grid_V], 1)
    grids_cuda = grid.requires_grad_(False).to(dtype=x.dtype, device=x.device)
    return grids_cuda


class WarpingLayer(nn.Module):
    def __init__(self):
        super(WarpingLayer, self).__init__()

    def forward(self, x, flow, height_im, width_im, div_flow):
        flo_list = []
        flo_w = flow[:, 0] * 2 / max(width_im - 1, 1) / div_flow
        flo_h = flow[:, 1] * 2 / max(height_im - 1, 1) / div_flow
        flo_list.append(flo_w)
        flo_list.append(flo_h)
        flow_for_grid = torch.stack(flo_list).transpose(0, 1)
        grid = torch.add(get_grid(x), flow_for_grid).transpose(1, 2).transpose(2, 3)        
        x_warp = F.grid_sample(x, grid, align_corners=True)

        mask = torch.ones(x.size(), requires_grad=False).to(dtype=x.dtype, device=x.device)
        mask = F.grid_sample(mask, grid, align_corners=True)
        mask = (mask >= 1.0).float()

        return x_warp * mask
